import _thread
import atexit
import contextlib
import datetime
import glob
import io
import json
import logging
import os
import threading
import time
from typing import Dict, List, Optional, Union

import ai2thor.controller
import ai2thor.server
import numpy as np
import typeguard  # can we replace with pydantic function validator?

logger = logging.getLogger(__name__)


# How far the player can reach.  I think this value needs to be bigger
# than the MAX_MOVE_DISTANCE or else the player may not be able to move
# into a position to reach some objects (it may be mathematically impossible).
DEFAULT_MOVE = 0.1

from .action import Action
from .config_manager import ConfigManager, SceneConfiguration
from .controller_events import (AfterStepPayload, BeforeStepPayload,
                                EndScenePayload, EventType, StartScenePayload)
from .controller_output_handler import ControllerOutputHandler
from .goal_metadata import GoalMetadata
from .parameter import Parameter, compare_param_values, rebuild_endhabituation
from .step_metadata import StepMetadata


def __reset_override(self, scene):
    # From https://github.com/allenai/ai2thor/blob/2.5.0/ai2thor/controller.py#L503-L525 # noqa: E501
    # Remove the error check: if scene not in self.scenes_in_build
    self.server.send(dict(action='Reset', sceneName=scene, sequenceId=0))
    self.last_event = self.server.receive()
    self.last_event = self.step(
        action='Initialize',
        **self.initialization_parameters
    )
    return self.last_event


ai2thor.controller.Controller.reset = __reset_override


def __image_depth_override(self, image_depth_data, **kwargs):
    # From https://github.com/NextCenturyCorporation/ai2thor/blob/47a9d0802861ba8d7a2a7a6d943a46db28ddbaab/ai2thor/server.py#L232-L240 # noqa: E501
    # The MCS depth shader in Unity is completely different now, so override
    # the original AI2-THOR depth image code. Just return what Unity sends us.
    return ai2thor.server.read_buffer_image(
        image_depth_data,
        self.screen_width,
        self.screen_height,
        dtype=np.float32
    )


ai2thor.server.Event._image_depth = __image_depth_override


class NumpyAwareEncoderOverride(json.JSONEncoder):
    # From https://github.com/allenai/ai2thor/blob/bd35d2cb887faee8b87aa04bd9373b027eb39f17/ai2thor/server.py#L17-L24 # noqa: E501
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.generic):
            return np.asscalar(obj)
        return super(NumpyAwareEncoderOverride, self).default(obj)


ai2thor.server.NumpyAwareEncoder = NumpyAwareEncoderOverride


class Controller():
    """
    MCS Controller class implementation for the MCS wrapper of the AI2-THOR
    library.

    https://ai2thor.allenai.org/ithor/documentation/

    Parameters
    ----------
    unity_app_file_path: str
    config: ConfigManager
    """

    @typeguard.typechecked
    def __init__(self, unity_app_file_path: str, config: ConfigManager):

        # Suppress print statements from the AI2-THOR Controller's constructor.
        with contextlib.redirect_stdout(io.StringIO()) as _:
            self._controller = ai2thor.controller.Controller(
                quality='Medium',
                fullscreen=False,
                headless=False,  # TODO confirm functionality
                local_executable_path=unity_app_file_path,
                width=config.get_screen_width(),
                height=config.get_screen_height(),
                scene='MCS',  # Unity scene name
                logs=True,
                # This constructor always initializes a scene, so add a scene
                # config to ensure it doesn't error
                sceneConfig={
                    "objects": []
                }
            )

        if not self._controller:
            raise Exception('AI2-THOR/Unity Controller failed to initialize')

        self._on_init()
        self._set_config(config)

    def _on_init(self):
        '''Set class variables after controller is initialized'''
        self._subscribers = []
        self._failure_handler_registered = False
        self._end_scene_called = False
        self._goal = GoalMetadata()
        self._last_step_check = -1
        self.__habituation_trial = 1
        # Output folder used to save debug image, video, and JSON files.
        self.__output_folder = None
        self._scene_config = None
        self.__step_number = 0
        self._timer = None
        self._timer_in_progress = False

    def _set_config(self, config: ConfigManager):
        '''Allows config to be changed without changing the controller and
        attached Unity process.  This typically should only be called by the
        MCS package itself.

        For users, call machine_common_sense.change_config()
        '''
        self._config = config
        self._output_handler = ControllerOutputHandler(self._config)
        self.parameter_converter = Parameter(config)

    def _check_step_for_timeout(self):
        '''Meant for use during eval, if a scene is hung on the same step
        for a period of time, end the scene.'''

        start_time = time.time()
        self._timer_in_progress = False
        timeout_seconds = self._config.get_timeout()

        if(self._last_step_check < self.__step_number):
            # skip step check for Initialize step
            if(self.__step_number != 0):
                self._last_step_check = self.__step_number

            timer_seconds = timeout_seconds - \
                ((time.time() - start_time) % timeout_seconds)
            self._timer = threading.Timer(timer_seconds,
                                          self._check_step_for_timeout)
            self._timer.daemon = True
            self._timer.start()
            self._timer_in_progress = True
        else:
            time_str = str(datetime.timedelta(seconds=timeout_seconds))
            logger.debug(
                f"Attempting to end scene due to inactivity (user not taking"
                f" any steps) for {time_str} (hh:mm:ss)")
            self.end_scene(rating=None, score=-1)

            _thread.interrupt_main()

    @typeguard.typechecked
    def start_scene(
        self, config_data: Union[SceneConfiguration, Dict]) \
            -> StepMetadata:
        """
        Starts a new scene using the given scene configuration data dict and
        returns the scene output data object.

        Parameters
        ----------
        config_data : SceneConfiguration or dict that can be serialized to
            SceneConfiguration
            The MCS scene configuration data for the scene to start.

        Returns
        -------
        StepMetadata
            The output data object from the start of the scene (the output from
            an "Initialize" action).
        """

        scene_config = self._convert_scene_config(config_data)

        self._scene_config = scene_config
        self.__habituation_trial = 1
        self.__step_number = 0
        self.__steps_in_lava = 0
        self.__triggered_by_sequence_incorrect = False
        self._goal = self._scene_config.retrieve_goal(
            self._config.get_steps_allowed_in_lava())
        self._end_scene_called = False

        skip_preview_phase = (scene_config.goal is not None and
                              scene_config.goal.skip_preview_phase)

        if (not self._scene_config.name):
            raise Exception('The `name` field in the scene '
                            'file cannot be empty.')

        if (self._config.is_file_writing_enabled()):
            os.makedirs(f'./{scene_config.name}', exist_ok=True)
            self.__output_folder = f'{os.getcwd()}/{scene_config.name}/'
            file_list = glob.glob(f'{self.__output_folder}*')
            for file_path in file_list:
                os.remove(file_path)

        sc = scene_config.dict(exclude_none=True, by_alias=True)

        ai2thor_step = self.parameter_converter.wrap_step(
            output_folder=self.__output_folder,
            action='Initialize',
            sceneConfig=sc)
        step_output = self._controller.step(ai2thor_step)

        self._output_handler.set_scene_config(scene_config)
        (pre_restrict_output, output) = self._output_handler.handle_output(
            step_output, self._goal, self.__step_number,
            self.__habituation_trial)

        self.__steps_in_lava = output.steps_on_lava
        self.__triggered_by_sequence_incorrect = \
            output.triggered_by_sequence_incorrect

        if not skip_preview_phase:
            if (self._goal is not None and
                    self._goal.last_preview_phase_step > 0):
                image_list = output.image_list
                depth_map_list = output.depth_map_list
                object_mask_list = output.object_mask_list

                logger.debug('STARTING PREVIEW PHASE...')

                for _ in range(self._goal.last_preview_phase_step):
                    output = self.step('Pass')
                    image_list = image_list + output.image_list
                    depth_map_list = depth_map_list + output.depth_map_list
                    object_mask_list = (object_mask_list +
                                        output.object_mask_list)

                logger.debug('ENDING PREVIEW PHASE')

                output.image_list = image_list
                output.depth_map_list = depth_map_list
                output.object_mask_list = object_mask_list

            logger.debug('NO PREVIEW PHASE')

            # TODO Should this be in the if block?  Now that we are using
            # subscribers, we may want to always register
            if(self._failure_handler_registered is False and
                    self._config.is_history_enabled()):
                # make sure history file is written when program exits
                atexit.register(self.end_scene, rating=None, score=-1)
                self._failure_handler_registered = True

        payload = self._create_post_step_event_payload_kwargs(
            ai2thor_step, step_output, pre_restrict_output, output)
        start_scene_payload = StartScenePayload(**payload)
        self._publish_event(
            EventType.ON_START_SCENE, start_scene_payload)

        self._check_step_for_timeout()

        return output

    def _convert_scene_config(self, config_data) -> SceneConfiguration:
        if isinstance(config_data, SceneConfiguration):
            return config_data
        return SceneConfiguration(**config_data)

    @typeguard.typechecked
    def step(self, action: str, **kwargs) -> Optional[StepMetadata]:
        """
        Runs the given action within the current scene.

        Parameters
        ----------
        action : string
            A selected action string from the list of available actions.
        **kwargs
            Zero or more key-and-value parameters for the action.

        Returns
        -------
        StepMetadata
            The MCS output data object from after the selected action and the
            physics simulation were run. Returns None if you have passed the
            "last_step" of this scene.

        Raises
        ------
            ValueError: If values are outside acceptable ranges or unable to
                convert to a number.
        """
        if (self._goal.last_step is not None and
                self._goal.last_step == self.__step_number):
            logger.error(
                "You have passed the last step for this scene. "
                "Ignoring your action. Please call controller.end_scene() "
                "now.")
            return None

        # if they call end scene action they should have
        #   called end_scene instead of step
        if action == Action.END_SCENE.value:
            self.end_scene()
            raise SystemExit(0)

        # reformulate hidden EndHabituation parameters
        if action == Action.END_HABITUATION.value:
            step_action_list = \
                self._goal._retrieve_unfiltered_action_list(self.__step_number)
            action = rebuild_endhabituation(step_action_list)

        if ',' in action:
            action, kwargs = Action.input_to_action_and_params(action)

        is_passive_scene = self._scene_config.is_passive_scene()
        action_list = self._goal.retrieve_action_list_at_step(
            self.__step_number, self.__steps_in_lava,
            self.__triggered_by_sequence_incorrect, is_passive_scene)

        # Only continue with this action step if the given action and
        # parameters are in the restricted action list.
        continue_with_step = any(action == restricted_action and (
            len(restricted_params.items()) == 0 or all(
                compare_param_values(restricted_params.get(key), value)
                for key, value in kwargs.items()
            )
        ) for restricted_action, restricted_params in action_list)

        if not continue_with_step:
            logger.error(
                f"The given action '{action}' with parameters "
                f"'{kwargs}' isn't in the action_list. Ignoring your action. "
                f"Possible actions at step {self.__step_number}:"
            )
            for action_data in action_list:
                logger.error(f'    {action_data}')
            raise ValueError(f"{action}-{kwargs} not in {action_list}")

        self.__step_number += 1

        payload = self._create_event_payload_kwargs()
        payload['action'] = action
        payload['habituation_trial'] = self.__habituation_trial
        payload['goal'] = self._goal

        self._publish_event(
            EventType.ON_BEFORE_STEP,
            BeforeStepPayload(**payload))

        if (action == Action.END_HABITUATION.value):
            self.__habituation_trial += 1

        if (self._goal.last_step is not None and
                self._goal.last_step == self.__step_number):
            logger.warning(
                "This is your last step for this scene. All "
                "your future actions will be skipped. Please call "
                "controller.end_scene() now.")

        ai2thor_step, params = self.parameter_converter.build_ai2thor_step(
            output_path=self.__output_folder,
            action=action,
            **kwargs)
        step_output = self._controller.step(ai2thor_step)

        (pre_restrict_output, output) = self._output_handler.handle_output(
            step_output, self._goal, self.__step_number,
            self.__habituation_trial)

        self.__steps_in_lava = output.steps_on_lava
        self.__triggered_by_sequence_incorrect = \
            output.triggered_by_sequence_incorrect

        payload = self._create_post_step_event_payload_kwargs(
            ai2thor_step, step_output, pre_restrict_output, output)
        payload['ai2thor_action'] = action
        payload['step_params'] = params
        payload['action_kwargs'] = kwargs
        self._publish_event(
            EventType.ON_AFTER_STEP,
            AfterStepPayload(**payload))

        return output

    @typeguard.typechecked
    def end_scene(
        self,
        rating: Optional[float] = None,
        score: Optional[float] = None,
        report: Dict[int, object] = None
    ) -> None:
        """
        Ends the current scene.  Calling end_scene() before calling
        start_scene() will do nothing. Calling end_scene() twice with
        the same scene will throw an exception.

        Parameters
        ----------
        rating : float, optional
            The plausibility rating to classify a passive / VoE scene as either
            plausible or implausible. Not used for any interactive scenes. For
            passive agent scenes, this rating should be continuous, from 0.0
            (completely implausible) to 1.0 (completely plausible). For other
            passive scenes, this rating must be binary, either 0 (implausible)
            or 1 (plausible). End-of-scene ratings are
            required for all passive / VoE scenes. (default None)
        score : float, optional
            The continuous plausibility score between 0.0 (completely
            implausible) and 1.0 (completely plausible). End-of-scene scores
            are required for all passive / VoE scenes except agent scenes.
            Not used for any interactive scenes or passive agent scenes.
            (default None)

            Note: when an issue causes the program to exit prematurely or
            end_scene isn't properly called but history_enabled is true,
            this value will be written to file as -1.
        report : Dict[int, object], optional
            Variable for retrospective per frame reporting for passive / VoE
            scenes. Not used for any interactive scenes or passive agent
            scenes. (default None)

            Key is an int representing a step/frame number from output step
            metadata, starting at 1. Value or payload contains:

                * rating : float or int, optional
                    The plausibility rating to classify a passive / VoE scene
                    as either plausible or implausible. Not used for any
                    interactive scenes. For passive agent scenes, this rating
                    should be continuous, from 0.0 (completely implausible) to
                    1.0 (completely plausible). For other passive scenes, this
                    rating must be binary, either 0 (implausible) or 1
                    (plausible). Frame-by-frame ratings are no
                    longer required for any scenes (but end-of-scene ratings
                    are). (default None)
                * score : float, optional
                    The continuous plausibility score between 0.0 (completely
                    implausible) and 1.0 (completely plausible). Frame-by-frame
                    scores are required for all passive / VoE scenes except
                    agent scenes. Not used for any interactive scenes or
                    passive agent scenes. (default None)
                * violations_xy_list : List[Dict[str, float]], optional
                    A list of one or more (x, y) locations
                    (ex: [{"x": 1, "y": 3.4}]), each representing a potential
                    violation-of-expectation. These locations are required for
                    all passive / VoE scenes except agent scenes. Not used for
                    any interactive scenes or passive agent scenes.
                    (default None)
                * internal_state : object, optional
                    A properly formatted json object representing various kinds
                    of internal states at a particular moment. Examples include
                    the estimated position of the agent, current map of the
                    world, etc. (default None)

            Example report:

            {

                    1: {

                        "rating": 1,

                        "score": 0.75,

                        "violations_xy_list": [{"x": 1,"y": 1}],

                        "internal_state": {"test": "some state"}

                    }

            }

        """
        if(not self._end_scene_called):
            payload = self._create_event_payload_kwargs()
            payload['rating'] = rating
            payload['score'] = score
            payload['report'] = report

            self._publish_event(
                EventType.ON_END_SCENE,
                EndScenePayload(
                    **payload))
            self._end_scene_called = True
        else:
            raise RuntimeError("end_scene called twice with the same scene")

        if (self._failure_handler_registered):
            atexit.unregister(self.end_scene)
            self._failure_handler_registered = False

        if (self._timer_in_progress):
            self._timer.cancel()
            self._timer_in_progress = False

    @typeguard.typechecked
    def stop_simulation(self) -> None:
        """Stop the 3D simulation environment. This controller won't work any
        more."""
        self._controller.stop()

    @typeguard.typechecked
    def get_metadata_level(self) -> str:
        """
        Returns the current metadata level set in the config. If none
        specified, returns 'default'.

        Returns
        -------
        string
            A string containing the current metadata level.
        """
        return self._config.get_metadata_tier().value

    @typeguard.typechecked
    def retrieve_object_states(self, object_id: str) -> List:
        """Return the state list at the current step for the object with the
        given ID from the scene configuration data, if any."""
        return self._scene_config.retrieve_object_states(
            object_id,
            self.__step_number)

    def subscribe(self, subscriber):
        if subscriber not in self._subscribers:
            self._subscribers.append(subscriber)

    def remove_all_event_handlers(self):
        self._subscribers = []

    @typeguard.typechecked
    def _publish_event(self, event_type: EventType,
                       payload: Union[StartScenePayload, BeforeStepPayload,
                                      AfterStepPayload,
                                      EndScenePayload]):
        for subscriber in self._subscribers:
            try:
                subscriber.on_event(event_type, payload)
            except Exception as msg:
                logger.error(
                    f"Error in event with type={event_type}"
                    f" to subscriber={type(subscriber)}",
                    exc_info=msg)

    def _create_event_payload_kwargs(self) -> dict:
        return {"step_number": self.__step_number,
                "config": self._config,
                "scene_config": self._scene_config}

    def _create_post_step_event_payload_kwargs(
            self, wrapped_step, step_metadata, step_output: StepMetadata,
            restricted_step_output: StepMetadata) -> dict:
        args = self._create_event_payload_kwargs()
        args['output_folder'] = self.__output_folder
        args['timestamp'] = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        args['wrapped_step'] = wrapped_step
        args['step_metadata'] = step_metadata
        args['step_output'] = step_output
        args['restricted_step_output'] = restricted_step_output
        args['goal'] = self._goal
        return args



import argparse

from runner_script import SingleFileRunnerScript


def action_callback(scene_data, step_metadata, runner_script):
    if runner_script.args.slow:
        if step_metadata.step_number < 360:
            if step_metadata.step_number % 10 == 0:
                return 'RotateRight', {}
            else:
                return 'Pass', {}
    else:
        if step_metadata.step_number < 36:
            return 'RotateRight', {}
    return None, None


class RotateRunnerScript(SingleFileRunnerScript):
    def _append_subclass_args_to_parser(
        self,
        parser: argparse.ArgumentParser
    ) -> argparse.ArgumentParser:
        parser = super()._append_subclass_args_to_parser(parser)
        parser.add_argument(
            '--slow',
            default=False,
            action='store_true',
            help='Rotate every 10 steps'
        )
        return parser


def main():
    RotateRunnerScript('Just Rotate', action_callback)


if __name__ == "__main__":
    main()



import textwrap
import unittest

import machine_common_sense as mcs


class TestObjectMetadata(unittest.TestCase):

    str_output = '''    {
        "uuid": "",
        "dimensions": [],
        "direction": {},
        "distance": -1.0,
        "distance_in_steps": -1.0,
        "distance_in_world": -1.0,
        "held": false,
        "mass": 0.0,
        "material_list": [],
        "position": {},
        "rotation": {},
        "segment_color": {},
        "shape": "",
        "state_list": [],
        "texture_color_list": [],
        "visible": false,
        "is_open": false,
        "openable": false,
        "locked": false,
        "associated_with_agent": "",
        "simulation_agent_held_object": "",
        "simulation_agent_is_holding_held_object": false
    }'''

    @classmethod
    def setUpClass(cls):
        cls.object_metadata = mcs.ObjectMetadata()

    @classmethod
    def tearDownClass(cls):
        # nothing to do
        pass

    def test_uuid(self):
        self.assertEqual(self.object_metadata.uuid, "")
        self.assertIsInstance(self.object_metadata.uuid, str)

    def test_associated_with_agent(self):
        self.assertEqual(self.object_metadata.associated_with_agent, "")
        self.assertIsInstance(self.object_metadata.associated_with_agent, str)

    def test_simulation_agent_held_object(self):
        self.assertEqual(self.object_metadata.simulation_agent_held_object, "")
        self.assertIsInstance(
            self.object_metadata.simulation_agent_held_object, str)

    def test_simulation_agent_is_holding_held_object(self):
        self.assertFalse(
            self.object_metadata.simulation_agent_is_holding_held_object)
        self.assertIsInstance(
            self.object_metadata.simulation_agent_is_holding_held_object, bool)

    def test_dimensions(self):
        self.assertFalse(self.object_metadata.dimensions)
        self.assertIsInstance(self.object_metadata.dimensions, list)

    def test_direction(self):
        self.assertFalse(self.object_metadata.direction)
        self.assertIsInstance(self.object_metadata.direction, dict)

    def test_distance(self):
        self.assertAlmostEqual(self.object_metadata.distance, -1.0)
        self.assertIsInstance(self.object_metadata.distance, float)

    def test_distance_in_steps(self):
        self.assertAlmostEqual(self.object_metadata.distance_in_steps, -1.0)
        self.assertIsInstance(self.object_metadata.distance_in_steps, float)

    def test_distance_in_world(self):
        self.assertAlmostEqual(self.object_metadata.distance_in_world, -1.0)
        self.assertIsInstance(self.object_metadata.distance_in_world, float)

    def test_held(self):
        self.assertFalse(self.object_metadata.held)
        self.assertIsInstance(self.object_metadata.held, bool)

    def test_mass(self):
        self.assertAlmostEqual(self.object_metadata.mass, 0.0)
        self.assertIsInstance(self.object_metadata.mass, float)

    def test_material_list(self):
        self.assertFalse(self.object_metadata.material_list)
        self.assertIsInstance(self.object_metadata.material_list, list)

    def test_position(self):
        self.assertFalse(self.object_metadata.position)
        self.assertIsInstance(self.object_metadata.position, dict)

    def test_rotation(self):
        self.assertFalse(self.object_metadata.rotation)
        self.assertIsInstance(self.object_metadata.rotation, dict)

    def test_segment_color(self):
        self.assertFalse(self.object_metadata.segment_color)
        self.assertIsInstance(self.object_metadata.segment_color, dict)

    def test_shape(self):
        self.assertEqual(self.object_metadata.shape, "")
        self.assertIsInstance(self.object_metadata.shape, str)

    def test_state_list(self):
        self.assertFalse(self.object_metadata.state_list)
        self.assertIsInstance(self.object_metadata.state_list, list)

    def test_texture_color_list(self):
        self.assertFalse(self.object_metadata.texture_color_list)
        self.assertIsInstance(self.object_metadata.texture_color_list, list)

    def test_visible(self):
        self.assertIsInstance(self.object_metadata.visible, bool)
        self.assertFalse(self.object_metadata.visible)

    def test_str(self):
        self.assertEqual(str(self.object_metadata),
                         textwrap.dedent(self.str_output))


if __name__ == '__main__':
    unittest.main()



from enum import Enum, unique
from typing import Tuple

import typeguard


@unique
class Action(Enum):
    """
    The actions available in the MCS simulation environment.

    For actions requiring objectImageCoords or receptacleObjectImageCoords,
    note that (0,0) represents the top left corner of the viewport, and that
    inputs must be greater than (0,0).
    """

    INITIALIZE = (
        "Initialize",
        "0",
        "Initialization"
    )
    """
    Initialize the scene. Intended only for internal use.
    """

    CLOSE_OBJECT = (
        "CloseObject",
        "1",
        "Close a nearby object. (objectId=string, amount=float "
        "(default:1), objectImageCoordsX=float, objectImageCoordsY=float)"
    )
    """
    Close a nearby object.

    Parameters
    ----------
    objectId : string, optional
        The "uuid" of the target object. Required unless the
        "objectImageCoords" properties are given.
    objectImageCoordsX : float, optional
        The X of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    objectImageCoordsY : float, optional
        The Y of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    amount : float
        The amount to close the object between 0 (completely opened) and 1
        (completely closed). Default: 1

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "IS_CLOSED_COMPLETELY"
        If the object is completely closed.
    "NOT_INTERACTABLE"
        If the object corresponding to the "objectImageCoords" vector is not an
        interactable object. This includes structural objects like the room's
        walls.
    "NOT_OBJECT"
        If the object corresponding to the "objectId" (or object corresponding
        to the "objectImageCoords" vector) is not an object.
    "NOT_OPENABLE"
        If the object itself cannot be closed.
    "NOT_RECEPTACLE"
        If the object corresponding to the "objectImageCoords" vector is not a
        receptacle object.
    "NOT_VISIBLE"
        If the object corresponding to the "objectId" is not in the viewport.
    "OUT_OF_REACH"
        If you cannot close the object because you are out of reach.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    DROP_OBJECT = (
        "DropObject",
        "2",
        "Drop an object you are holding. (objectId=string)"
    )
    """
    Drop an object you are holding.

    Parameters
    ----------
    objectId : string, optional
        The "uuid" of the held object. Defaults to the first held object.

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "NOT_HELD"
        If you cannot put down the object corresponding to the "objectId"
        because you are not holding it.
    "NOT_OBJECT"
        If the object corresponding to the "objectId" is not an object.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    MOVE_AHEAD = (
        "MoveAhead",
        "w",
        "Move yourself ahead based on your current view."
    )
    """
    Move yourself forward based on your current viewport.


    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "OBSTRUCTED"
        If you cannot move forward because your path is obstructed.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    MOVE_BACK = (
        "MoveBack",
        "s",
        "Move yourself back based on your current view."
    )
    """
    Move yourself backward based on your current viewport.

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "OBSTRUCTED"
        If you cannot move backward because your path is obstructed.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    MOVE_LEFT = (
        "MoveLeft",
        "a",
        "Move yourself to your left based on your current view."
    )
    """
    Move yourself left based on your current viewport.

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "OBSTRUCTED"
        If you cannot move left because your path is obstructed.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    MOVE_RIGHT = (
        "MoveRight",
        "d",
        "Move yourself to your right based on your current view."
    )
    """
    Move yourself right based on your current viewport.

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "OBSTRUCTED"
        If you cannot move right because your path is obstructed.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    OPEN_OBJECT = (
        "OpenObject",
        "3",
        "Open a nearby object. (objectId=string, "
        "amount=float (default:1), objectImageCoordsX=float, "
        "objectImageCoordsY=float)"
    )
    """
    Open a nearby object.

    Parameters
    ----------
    objectId : string, optional
        The "uuid" of the target object. Required unless the
        "objectImageCoords" properties are given.
    objectImageCoordsX : float, optional
        The X of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    objectImageCoordsY : float, optional
        The Y of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    amount : float
        The amount to open the object between 0 (completely closed) and 1
        (completely opened). Default: 1

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "IS_OPENED_COMPLETELY"
        If the object is completely opened.
    "NOT_INTERACTABLE"
        If the object corresponding to the "objectImageCoords" vector is not an
        interactable object. This includes structural objects like the room's
        walls.
    "NOT_OBJECT"
        If the object corresponding to the "objectId" (or object corresponding
        to the "objectImageCoords" vector) is not an object.
    "NOT_OPENABLE"
        If the object itself cannot be opened.
    "NOT_RECEPTACLE"
        If the object corresponding to the "objectImageCoords" vector is not a
        receptacle object.
    "NOT_VISIBLE"
        If the object corresponding to the "objectId" is not in the viewport.
    "OBSTRUCTED"
        If you cannot open the object because you will be in the way of the
        object when its opened.
    "OUT_OF_REACH"
        If you cannot open the object because you are out of reach.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    PICKUP_OBJECT = (
        "PickupObject",
        "4",
        "Pickup a nearby object and hold it in your hand. "
        "(objectId=string, objectImageCoordsX=float, objectImageCoordsY=float)"
    )
    """
    Pick up a nearby object and hold it in your hand. This action incorporates
    reaching out your hand in front of you, opening your fingers, and grabbing
    the object. You may hold multiple objects simultaneously.

    Parameters
    ----------
    objectId : string, optional
        The "uuid" of the target object. Required unless the
        "objectImageCoords" properties are given.
    objectImageCoordsX : float, optional
        The X of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    objectImageCoordsY : float, optional
        The Y of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "NOT_INTERACTABLE"
        If the object corresponding to the "objectImageCoords" vector is not an
        interactable object.
    "NOT_OBJECT"
        If the object corresponding to the "objectId" (or object corresponding
        to the "objectImageCoords" vector) is not an object.
    "NOT_PICKUPABLE"
        If the object itself cannot be picked up.
    "NOT_VISIBLE"
        If the object corresponding to the "objectId" is not in the viewport.
    "OUT_OF_REACH"
        If you cannot pick up the object because you are out of reach.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    PULL_OBJECT = (
        "PullObject",
        "5",
        "Pull a nearby object. (objectId=string, force=float (default:0.5), "
        "objectImageCoordsX=float, objectImageCoordsY=float)"
    )
    """
    Pull a nearby object by applying a physical force directly toward you on
    the X/Z axis to the center point of the object (note that this means it
    does not matter where you Pull on the object, since the force is always
    applied to the center point).

    Parameters
    ----------
    objectId : string, optional
        The "uuid" of the target object. Required unless the
        "objectImageCoords" properties are given.
    objectImageCoordsX : float, optional
        The X of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    objectImageCoordsY : float, optional
        The Y of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    force : float
        The amount of force, from 0 to 1, used to move the target object.
        Default: 0.5

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "NOT_INTERACTABLE"
        If the object corresponding to the "objectImageCoords" vector is not an
        interactable object.
    "NOT_OBJECT"
        If the object corresponding to the "objectId" (or object corresponding
        to the "objectImageCoords" vector) is not an object.
    "NOT_MOVEABLE"
        If the object itself cannot be moved by a baby.
    "NOT_VISIBLE"
        If the object corresponding to the "objectId" is not in the viewport.
    "OUT_OF_REACH"
        If you cannot move the object because you are out of reach.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    PUSH_OBJECT = (
        "PushObject",
        "6",
        "Push a nearby object. (objectId=string, force=float (default:0.5), "
        "objectImageCoordsX=float, objectImageCoordsY=float)"
    )
    """
    Push a nearby object by applying a physical force directly away from you on
    the X/Z axis to the center point of the object (note that this means it
    does not matter where you Push on the object, since the force is always
    applied to the center point).

    Parameters
    ----------
    objectId : string, optional
        The "uuid" of the target object. Required unless the
        "objectImageCoords" properties are given.
    objectImageCoordsX : float, optional
        The X of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    objectImageCoordsY : float, optional
        The Y of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    force : float
        The amount of force, from 0 to 1, used to move the target object.
        Default: 0.5

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "NOT_INTERACTABLE"
        If the object corresponding to the "objectImageCoords" vector is not an
        interactable object.
    "NOT_OBJECT"
        If the object corresponding to the "objectId" (or object corresponding
        to the "objectImageCoords" vector) is not an object.
    "NOT_MOVEABLE"
        If the object itself cannot be moved by a baby.
    "NOT_VISIBLE"
        If the object corresponding to the "objectId" is not in the viewport.
    "OUT_OF_REACH"
        If you cannot move the object because you are out of reach.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    PUT_OBJECT = (
        "PutObject",
        "7",
        "Place an object you are holding into/onto a nearby "
        "receptacle object. (objectId=string, receptacleObjectId=string, "
        "receptacleObjectImageCoordsX=float, "
        "receptacleObjectImageCoordsY=float)"
    )
    """
    Put down an object you are holding into/onto a nearby receptacle object. A
    receptacle is an object that can hold other objects, like a block, box,
    drawer, shelf, or table.

    Parameters
    ----------
    objectId : string, optional
        The "uuid" of the held object. Defaults to the first held object.
    receptacleObjectId : string, optional
        The "uuid" of the target receptacle. Required unless the
        "receptacleObjectImageCoords" properties are given.
    receptacleObjectImageCoordsX : float, optional
        The X of a pixel coordinate on the target receptacle based
        on your current viewport. Can be used in place of the
        "receptacleObjectId" property.
        (See note under "Action" header regarding image coordinates.)
    receptacleObjectImageCoordsY : float, optional
        The Y of a pixel coordinate on the target receptacle based
        on your current viewport. Can be used in place of the
        "receptacleObjectId" property.
        (See note under "Action" header regarding image coordinates.)

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "NOT_HELD"
        If you cannot put down the object corresponding to the "objectId"
        because you are not holding it.
    "NOT_INTERACTABLE"
        If the object corresponding to the "objectImageCoords" or
        "receptacleObjectImageCoords" vector is not an interactable object.
        This includes structural objects like the room's walls.
    "NOT_OBJECT"
        If the object corresponding to the "objectId" and/or
        "receptacleObjectId" (or object corresponding to the
        "receptacleObjectImageCoords" vector) is not an object.
    "NOT_RECEPTACLE"
        If the object corresponding to the "receptacleObjectId" (or object
        corresponding to the "receptacleObjectImageCoords" vector) is not a
        receptacle.
    "NOT_VISIBLE"
        If the object corresponding to the "objectId" is not in the viewport.
    "OBSTRUCTED"
        If you cannot put down the object because your path is obstructed.
    "OUT_OF_REACH"
        If you cannot put down the object because you are out of reach.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    TORQUE_OBJECT = (
        "TorqueObject",
        "8",
        "Apply torque to a nearby object. (objectId=string, "
        "force=float(default:0.5), objectImageCoordsX=float, "
        "objectImageCoordsY=float)"
    )
    """
    Apply torque to a nearby object.

    Parameters
    ----------
    objectId : string, optional
        The "uuid" of the target object. Required unless the
        "objectImageCoords" properties are given.
    objectImageCoordsX : float, optional
        The X of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    objectImageCoordsY : float, optional
        The Y of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    force : float
        The amount of force, from -1 to 1, used to move the target object.
        Default: 0.5

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "NOT_INTERACTABLE"
        If the object corresponding to the "objectImageCoords" vector is not an
        interactable object.
    "NOT_OBJECT"
        If the object corresponding to the "objectId" (or object corresponding
        to the "objectImageCoords" vector) is not an object.
    "NOT_MOVEABLE"
        If the object itself cannot be moved by a baby.
    "NOT_VISIBLE"
        If the object corresponding to the "objectId" is not in the viewport.
    "OUT_OF_REACH"
        If you cannot move the object because you are out of reach.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    ROTATE_OBJECT = (
        "RotateObject",
        "9",
        "Apply a rotation of 5 degrees to a nearby object. "
        "(objectId=string, "
        "clockwise=bool(default:True), objectImageCoordsX=float, "
        "objectImageCoordsY=float)"
    )
    """
    Apply a rotation of 5 degrees to a nearby object. Will fail if rotating
    the object would cause it to collide with another object or the performer
    agent, returning OBSTRUCTED.

    Parameters
    ----------
    objectId : string, optional
        The "uuid" of the target object. Required unless the
        "objectImageCoords" properties are given.
    objectImageCoordsX : float, optional
        The X of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    objectImageCoordsY : float, optional
        The Y of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    clockwise : bool
        If the rotation should be clockwise.
        Default: True

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "NOT_INTERACTABLE"
        If the object corresponding to the "objectImageCoords" vector is not an
        interactable object.
    "NOT_OBJECT"
        If the object corresponding to the "objectId" (or object corresponding
        to the "objectImageCoords" vector) is not an object.
    "NOT_MOVEABLE"
        If the object itself cannot be moved by a baby.
    "NOT_VISIBLE"
        If the object corresponding to the "objectId" is not in the viewport.
    "OBSTRUCTED"
        If you cannot rotate the object because the path of rotation
        is obstructed.
    "OUT_OF_REACH"
        If you cannot move the object because you are out of reach.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    MOVE_OBJECT = (
        "MoveObject",
        "m",
        "Apply a movement of 0.1 meters to a nearby object. "
        "(objectId=string, "
        "lateral=int(default:0), "
        "straight=int(default:1), "
        "objectImageCoordsX=float, "
        "objectImageCoordsY=float)"
    )
    """
    Apply a movement of 0.1 meters units to a nearby object. If the object
    would come into contact with another object, and the other object is small
    and moveable, the other object will also move 0.1 meters in the same
    direction. This movement does not attempt to simulate realistic physics
    in regard to collisions with other object(s). If you wish to simulate
    realistic physical movement, please use the PullObject and PushObject
    actions instead, which apply a force using the environment's physics
    simulation engine.

    Parameters
    ----------
    objectId : string, optional
        The "uuid" of the target object. Required unless the
        "objectImageCoords" properties are given.
    objectImageCoordsX : float, optional
        The X of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    objectImageCoordsY : float, optional
        The Y of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    lateral : int
        The x axis direction of movement on the object relative to the agent.
        Can be -1, 0, 1. If only lateral is given,
        straight will default to 0
        Default: 0
    straight : int
        The x axis direction of movement on the object relative to the agent.
        Can be -1, 0, 1. If only straight is given,
        lateral will default to 0
        Default: 1

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "NOT_INTERACTABLE"
        If the object corresponding to the "objectImageCoords" vector is not an
        interactable object.
    "NOT_OBJECT"
        If the object corresponding to the "objectId" (or object corresponding
        to the "objectImageCoords" vector) is not an object.
    "NOT_MOVEABLE"
        If the object itself cannot be moved by a baby.
    "NOT_VISIBLE"
        If the object corresponding to the "objectId" is not in the viewport.
    "OBSTRUCTED"
        If you cannot rotate the object because the path of movement
        is obstructed.
    "OUT_OF_REACH"
        If you cannot move the object because you are out of reach.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    INTERACT_WITH_AGENT = (
        "InteractWithAgent",
        "t",
        "Interact with an agent. If that agent has an object, "
        "it will hold out the object for you to pickup; "
        "otherwise, the agent will look sad."
        "(objectId=string, "
        "objectImageCoordsX=float, "
        "objectImageCoordsY=float)"
    )
    """
    Interact with an agent. If that agent has an object, it will hold
    out the object for you to pickup; otherwise, the agent will look sad.
    If the agent was pointing at an object, the agent will resume pointing
    afterward.

    Parameters
    ----------
    objectId : string, optional
        The "uuid" of the target object. Required unless the
        "objectImageCoords" properties are given.
    objectImageCoordsX : float, optional
        The X of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)
    objectImageCoordsY : float, optional
        The Y of a pixel coordinate on the target object based on
        your current viewport. Can be used in place of the "objectId" property.
        (See note under "Action" header regarding image coordinates.)

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "NOT_INTERACTABLE"
        If the object corresponding to the "objectImageCoords" vector is not an
        interactable object.
    "NOT_OBJECT"
        If the object corresponding to the "objectId" (or object corresponding
        to the "objectImageCoords" vector) is not an object.
    "NOT_AGENT"
        If the object being interacted with is not a simulation agent
    "AGENT_CURRENTLY_INTERACTING_WTIH_PERFORMER"
        If the object being interacted with is a simulation agent already
        interacting with the performer.
    "AGENT_IS_BUSY"
        If the object being interacted with is a simulation agent that is
        currently rotating to face an object or beginning its point animation.
    "NOT_VISIBLE"
        If the object corresponding to the "objectId" is not in the viewport.
    "OUT_OF_REACH"
        If you cannot move the object because you are out of reach.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    LOOK_UP = (
        "LookUp",
        "i",
        "Rotate your view up (subtract 10 degrees from head tilt)."
    )
    """
    Rotate your view up (subtract 10 degrees from head tilt).

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "CANNOT_ROTATE"
        Failed because you cannot look down/up more than +/- 90 degrees.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    LOOK_DOWN = (
        "LookDown",
        "k",
        "Rotate your view down (add 10 degrees to head tilt)."
    )
    """
    Rotate your viewport down (add 10 degrees to head tilt).

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "CANNOT_ROTATE"
        Failed because you cannot look down/up more than +/- 90 degrees.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    ROTATE_LEFT = (
        "RotateLeft",
        "j",
        "Rotate your view left by 10 degrees."
    )
    """
    Rotate your viewport left by 10 degrees.

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    ROTATE_RIGHT = (
        "RotateRight",
        "l",
        "Rotate your view right by 10 degrees."
    )
    """
    Rotate your viewport right by 10 degrees.

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    END_HABITUATION = (
        "EndHabituation",
        "h",
        "Ends a habituation trial for the scene by blanking the screen "
        "for one action (and teleporting the agent if needed). Sometimes"
        " needed depending on the task type."
    )
    """
    Ends a habituation trial for the scene by blanking the screen for one
    action (and teleporting the agent if needed). Sometimes needed depending
    on the task type.

    Note that we currently plan to use the starting position/rotation as
    teleport parameters here for applicable cases. We cannot currently
    guarantee that using a position intersecting another object or outside
    the room won't cause issues or errors.

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    # Pass should always be the last action in the enum.
    END_SCENE = (
        "EndScene",
        " ",
        "There is no action available and end_scene needs to be called."
    )
    """
    Call end_scene now there is no actions available.
    Does nothing.

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    # Pass should always be the last action in the enum.
    PASS = (
        "Pass",
        " ",
        "Do nothing. (no params)"
    )
    """
    Do nothing.

    Returns
    -------
    "SUCCESSFUL"
        Action successful.
    "FAILED"
        Unexpected error; please report immediately to development team.
    """

    def __new__(cls, action, key, desc):
        obj = object.__new__(cls)
        obj._value_ = action
        obj._key = key
        obj._desc = desc
        # The following line is to get enum name from more than one value:
        # https://stackoverflow.com/a/43210118/6997391
        cls._value2member_map_[key] = obj
        return obj

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}.{self._name_}: "
            f"{', '.join([self._value_, self._key, self._desc])}>")

    @property
    def key(self):
        return self._key

    @property
    def desc(self):
        return self._desc

    @staticmethod
    @typeguard.typechecked
    def input_to_action_and_params(input_str: str) -> Tuple:
        """
        Transforms the given input string into an action string
        and parameter dict.

        Parameters
        ----------
        input_value : string
            The input value.

        Returns
        -------
        string
            The action string, or None if the given input had an error
            transforming the action string.
        dict
            The parameter dict, or None if the given input had an error
            transforming parameters.
        """
        input_split = input_str.split(',')
        action = input_split[0]

        try:
            _ = Action(action).name
        except BaseException:
            return None, {}

        if len(input_split) < 2:
            return action, {}

        params = {}

        try:
            for param in input_split[1:]:
                param_key, param_value = param.split('=')
                params[param_key.strip()] = param_value.strip()

                try:
                    params[param_key.strip()] = float(param_value.strip())
                except ValueError:
                    pass

        except BaseException:
            return action, None

        return action, params


FORCE_ACTIONS = [
    Action.PUSH_OBJECT,
    Action.PULL_OBJECT,
    Action.TORQUE_OBJECT]
OBJECT_MOVE_ACTIONS = [
    Action.CLOSE_OBJECT,
    Action.OPEN_OBJECT]
MOVE_ACTIONS = [
    Action.MOVE_AHEAD,
    Action.MOVE_LEFT,
    Action.MOVE_RIGHT,
    Action.MOVE_BACK]
OBJECT_IMAGE_ACTIONS = [
    Action.CLOSE_OBJECT,
    Action.OPEN_OBJECT,
    Action.PICKUP_OBJECT,
    Action.PUSH_OBJECT,
    Action.PULL_OBJECT,
    Action.TORQUE_OBJECT,
    Action.ROTATE_OBJECT,
    Action.INTERACT_WITH_AGENT,
    Action.MOVE_OBJECT]
RECEPTACLE_ACTIONS = [
    Action.PUT_OBJECT]



import argparse
import cmd

import machine_common_sense as mcs
from machine_common_sense.action import (FORCE_ACTIONS, OBJECT_IMAGE_ACTIONS,
                                         OBJECT_MOVE_ACTIONS,
                                         RECEPTACLE_ACTIONS, Action)
from machine_common_sense.goal_metadata import GoalMetadata
from machine_common_sense.logging_config import LoggingConfig

try:
    from getch_helper import getch
except ImportError:
    from .getch_helper import getch  # noqa: F401

commands = []


def parse_args():
    parser = argparse.ArgumentParser(description='Run MCS')
    parser.add_argument(
        'mcs_scene_json_file',
        help='MCS JSON scene configuration file to load')
    parser.add_argument(
        '--mcs_unity_build_file',
        type=str,
        default=None,
        help='Path to MCS unity build file')
    parser.add_argument(
        '--mcs_unity_version',
        type=str,
        default=None,
        help='version of MCS Unity executable.  Default: current')
    parser.add_argument(
        '--config_file_path',
        type=str,
        default=None,
        required=True,
        help='Path to configuration file to read in and set various '
        'properties, such as metadata level and whether or not to '
        'save history files properties.')
    return parser.parse_args()


class Command:
    '''Class to contain possible commands and keys.'''

    def __init__(self, name, key, desc):
        self.name = name
        self.key = key
        self.desc = desc


class HumanInputShell(cmd.Cmd):

    prompt = '(command)->'

    def __init__(self, input_controller, input_scene_data):
        super(HumanInputShell, self).__init__()

        self.controller = input_controller
        self.scene_data = input_scene_data
        self.previous_output = None
        self.auto = False

    def precmd(self, line):
        return line

    def postcmd(self, stop_flag, line) -> bool:
        print('================================================='
              '==============================')
        return stop_flag

    def default(self, line):
        split_input = line.split(',')

        # Check for shortcut key, if attempted shortcut key, map and check
        # valid key
        try:
            if len(split_input[0]) == 1:
                split_input[0] = mcs.Action(split_input[0]).value
        except BaseException:
            print(
                "You entered an invalid shortcut key, please try again. "
                "(Type 'help' to display commands again)")
            print(f"You entered: {split_input[0]}")
            return

        if (
            self.auto and self.previous_output.action_list and
            len(self.previous_output.action_list) == 1
        ):
            print('Automatically selecting the only available action...')
            action, params = self.previous_output.action_list[0]
        else:
            action, params = mcs.Action.input_to_action_and_params(
                ','.join(split_input)
            )

        if action is None:
            print(
                f"You entered an invalid command, '{split_input[0]}', please "
                f"try again.  (Type 'help' to display commands again)")
            return

        if params is None:
            print(
                "ERROR: Parameters should be separated by commas, and "
                "look like this example: rotation=45")
            return

        params = self._is_parameters_valid(action, params)
        output = self.controller.step(action, **params)

        # The output may be None if given an invalid action.
        if output:
            self.previous_output = output
            self.show_available_actions(self.previous_output)
            # If auto mode is True, loop as long as possible.
            if (
                self.auto and self.previous_output.action_list and
                len(self.previous_output.action_list) == 1
            ):
                self.default('')

    def _set_default_image_coords(self, string, action_str, params):
        number_of_default_img_coords_to_use = (
            1 if f'{string}ImageCoordsX' in params or
            f'{string}ImageCoordsY' in params else 2)
        obj_img_coord_x = (
            300 if f'{string}ImageCoordsX' not in
            params else params[f'{string}ImageCoordsX'])
        obj_img_coord_y = (
            200 if f'{string}ImageCoordsY' not in
            params else params[f'{string}ImageCoordsY'])
        print(
            f"Action: '{action_str}' using "
            f"{number_of_default_img_coords_to_use} default image "
            f"coord{'s' if number_of_default_img_coords_to_use == 0 else ''} "
            f"x={obj_img_coord_x} y={obj_img_coord_y} because no {string}Id "
            f"or {string}ImageCoords were given")
        params[f'{string}ImageCoordsX'] = obj_img_coord_x
        params[f'{string}ImageCoordsY'] = obj_img_coord_y
        return params

    def _is_parameters_valid(self, action_str, params):
        action: Action = Action(action_str)
        valid = False
        if action in (FORCE_ACTIONS + OBJECT_IMAGE_ACTIONS +
                      OBJECT_MOVE_ACTIONS):
            valid = ('objectId' in params or (
                'objectImageCoordsX' in params and
                'objectImageCoordsY' in params))
            if not valid:
                params = self._set_default_image_coords(
                    "object", action_str, params)
        elif action in RECEPTACLE_ACTIONS:
            valid = ('receptacleObjectId' in params or (
                'receptacleObjectImageCoordsX' in params and
                'receptacleObjectImageCoordsY' in params))
            if not valid:
                params = self._set_default_image_coords(
                    "receptacleObject", action_str, params)
        return params

    def help_auto(self):
        print(
            "Automatically runs the next action if only one action is "
            "available at that step."
        )

    def help_print(self):
        print("Prints all commands that the user can use.")

    def help_exit(self):
        print("Exits out of the MCS Human Input program.")

    def help_reset(self):
        print("Resets the scene.")

    def help_shortcut_key_mode(self):
        print("Toggles on mode where the user can execute single key "
              "commands without hitting enter.")

    def do_auto(self, args=None):
        self.auto = (not self.auto)
        print(f"Toggle auto mode {('ON' if self.auto else 'OFF')}")
        if (
            self.auto and self.previous_output.action_list and
            len(self.previous_output.action_list) == 1
        ):
            self.default('')

    def do_exit(self, args=None) -> bool:
        print("Exiting Human Input Mode\n")
        self.controller.end_scene()
        return True

    def do_print(self, args=None):
        print_commands()

    def do_reset(self, args=None):
        print("Resetting the current scene...")
        self.previous_output = (self.controller).start_scene(self.scene_data)
        self.show_available_actions(self.previous_output)

    def do_shortcut_key_mode(self, args=None):
        print("Entering shortcut mode...")
        print("Press key 'e' to exit\n")
        list_of_action_keys = [
            action.key for action in mcs.Action]

        while True:
            char = getch.__call__()
            print('\n(shortcut-command)->', char)
            if char == 'e':  # exit shortcut key mode
                break
            elif char in list_of_action_keys:
                self.default(char)

    def show_available_actions(self, output):
        if (
            output.action_list and
            len(output.action_list) < len(GoalMetadata.DEFAULT_ACTIONS)
        ):
            print('Only actions available during this step:')
            for action, params in output.action_list:
                action_string = action + ''.join(
                    f',{key}={value}' for key, value in params.items()
                )

                print(f'    {action_string}')
        else:
            print('All actions available during this step.')


def build_commands():
    '''Define all the possible human input commands.'''
    for action in mcs.Action:
        commands.append(Command(action.value,
                                action.key,
                                action.desc))


def print_commands():
    '''Display all the possible commands and key mappings to the user.'''
    print(" ")
    print("--------------- Available Commands ---------------")
    print(" ")

    for command in commands:
        print(
            f"- {command.name} (ShortcutKey={command.key}): {command.desc}")

    print(" ")
    print("---------------- Example Commands ----------------")
    print(" ")
    print("MoveAhead")
    print("RotateLeft")
    print("LookUp")
    print(" ")
    print("----------------- Other Commands -----------------")
    print(" ")
    print("Enter 'auto' to start auto mode.")
    print("Enter 'print' to print the commands again.")
    print("Enter 'reset' to reset the scene.")
    print("Enter 'exit' to exit the program.")
    print("Enter 'shortcut_key_mode' to be able to enter Commands "
          "without hitting enter.")
    print(" ")
    print("----------------- Features -----------------------")
    print("Up/Down arrow keys go through the command history")
    print("Tab key is for auto completion of a command")
    print("------------------ End Commands ------------------")
    print(" ")


def run_scene(controller, scene_data):
    '''Run scene loaded in the scene config data.'''
    build_commands()
    print_commands()

    input_commands = HumanInputShell(controller, scene_data)
    input_commands.do_reset()
    input_commands.cmdloop()


def main():
    mcs.init_logging(LoggingConfig.get_dev_logging_config())
    args = parse_args()
    scene_data = mcs.load_scene_json_file(args.mcs_scene_json_file)

    controller = mcs.create_controller(
        unity_app_file_path=args.mcs_unity_build_file,
        config_file_or_dict=args.config_file_path,
        unity_cache_version=args.mcs_unity_version)

    scene_file_path = args.mcs_scene_json_file
    scene_file_name = scene_file_path[scene_file_path.rfind('/') + 1:]

    if 'name' not in scene_data.keys():
        scene_data['name'] = scene_file_name[0:scene_file_name.find('.')]

    if controller is not None:
        run_scene(controller, scene_data)


if __name__ == "__main__":
    main()



import os
import unittest
from unittest.mock import DEFAULT, patch

import machine_common_sense as mcs
from machine_common_sense.config_manager import (ActionConfig,
                                                 AgentMovementConfig,
                                                 AgentSettings,
                                                 ChangeMaterialConfig,
                                                 ConfigManager, ForceConfig,
                                                 Goal, LidConfig, MetadataTier,
                                                 MoveConfig, OpenCloseConfig,
                                                 PhysicsConfig,
                                                 SceneConfiguration,
                                                 SceneObject, SequenceConfig,
                                                 ShowConfig, SingleStepConfig,
                                                 SizeConfig,
                                                 StepBeginEndConfig,
                                                 TeleportConfig,
                                                 TransformConfig, Vector3d)


class TestConfigManager(unittest.TestCase):

    def mock_env(**env_vars):
        return patch.dict(os.environ, env_vars, clear=True)

    @classmethod
    @mock_env()
    def setUpClass(cls):
        cls.config_mngr = ConfigManager(config_file_or_dict={})
        cls.config_mngr._config[
            cls.config_mngr.CONFIG_DEFAULT_SECTION
        ] = {}

    @ classmethod
    def tearDownClass(cls):
        pass

    def test_init(self):
        self.assertIsNotNone(self.config_mngr._config)

    @ mock_env()
    @ patch.multiple(ConfigManager, _read_in_config_dict=DEFAULT,
                     _read_in_config_file=DEFAULT)
    def test_init_with_arg(self, _read_in_config_dict, _read_in_config_file):
        file_path = './arg-test.ini'
        _ = ConfigManager(file_path)

        _read_in_config_file.assert_called_once_with(file_path)
        _read_in_config_dict.assert_not_called()

    @ mock_env(MCS_CONFIG_FILE_PATH='~/somefolder/env-var-test.ini')
    @ patch.multiple(ConfigManager, _read_in_config_dict=DEFAULT,
                     _read_in_config_file=DEFAULT)
    def test_init_with_env_variable(
            self, _read_in_config_dict, _read_in_config_file):
        _ = ConfigManager()

        _read_in_config_dict.assert_not_called()
        _read_in_config_file.assert_called_once_with(
            os.environ.get('MCS_CONFIG_FILE_PATH'))

    @mock_env(MCS_CONFIG_FILE_PATH='~/somefolder/env-var-test.ini')
    @patch.multiple(ConfigManager, _read_in_config_dict=DEFAULT,
                    _read_in_config_file=DEFAULT)
    def test_init_with_arg_and_env_variable(
            self, _read_in_config_dict, _read_in_config_file):
        '''Ensures environment variable overrides provided file path'''
        file_path = './arg-test.ini'
        _ = ConfigManager(file_path)

        _read_in_config_dict.assert_not_called()
        _read_in_config_file.assert_called_once_with(
            os.environ.get('MCS_CONFIG_FILE_PATH')
        )

    @mock_env(MCS_CONFIG_FILE_PATH='~/somefolder/env-var-test.ini')
    def test_nonexistent_file_from_env_variable(self):
        '''Missing config files should raise an exception'''
        with self.assertRaises(RuntimeError):
            ConfigManager()

    @mock_env(MCS_CONFIG_FILE_PATH=('machine_common_sense/scripts/'
                                    'config_level2_debug.ini'))
    def test_init_no_override_with_env_var_and_dict(self):
        config_options = {
            'metadata': 'oracle'
        }
        config_mngr = ConfigManager(config_options)
        self.assertEqual(config_mngr.get_metadata_tier(),
                         MetadataTier.LEVEL_2)

    @mock_env(MCS_CONFIG_FILE_PATH=('machine_common_sense/scripts/'
                                    'config_level2_debug.ini'))
    @patch.multiple(ConfigManager, _read_in_config_dict=DEFAULT,
                    _read_in_config_file=DEFAULT)
    def test_init_env_var_and_dict_function_calls(
            self, _read_in_config_dict, _read_in_config_file):
        config_options = {
            'metadata': 'oracle'
        }
        _ = ConfigManager(config_options)
        _read_in_config_dict.assert_not_called()
        _read_in_config_file.assert_called_once_with(
            os.environ.get('MCS_CONFIG_FILE_PATH'))

    def test_init_with_filepath(self):
        config_file = 'machine_common_sense/scripts/config_level2_debug.ini'
        config_mngr = ConfigManager(config_file_or_dict=config_file)
        self.assertEqual(config_mngr.get_metadata_tier(),
                         MetadataTier.LEVEL_2)

    def test_init_with_filepath_missing(self):
        '''Provided config file path must exist or an exception occurs'''
        missing_config_file = '~/somefolder/env-var-test.ini'
        with self.assertRaises(RuntimeError):
            ConfigManager(missing_config_file)

    def test_init_with_no_config(self):
        '''No dictionary, no file, and no env should raise an exception'''
        with self.assertRaises(RuntimeError):
            ConfigManager()

    def test_init_with_dict(self):
        config_options = {
            'metadata': 'oracle',
            'video_enabled': 'false',
            'save_debug_images': True
        }

        config_mngr = ConfigManager(config_options)

        self.assertIsNotNone(config_mngr._config)
        self.assertEqual(config_mngr.get_metadata_tier(),
                         MetadataTier.ORACLE)
        self.assertFalse(config_mngr.is_video_enabled())
        self.assertTrue(config_mngr.is_save_debug_images())

    def test_validate_screen_size(self):
        self.config_mngr._config[
            self.config_mngr.CONFIG_DEFAULT_SECTION
        ][
            self.config_mngr.CONFIG_SIZE
        ] = '450'
        self.config_mngr._validate_screen_size()

        self.assertEqual(self.config_mngr.get_size(), 450)

        self.config_mngr._config[
            self.config_mngr.CONFIG_DEFAULT_SECTION
        ][
            self.config_mngr.CONFIG_SIZE
        ] = '449'
        self.config_mngr._validate_screen_size()

        self.assertEqual(self.config_mngr.get_size(),
                         self.config_mngr.SCREEN_WIDTH_DEFAULT)

    def test_get_evaluation_name(self):
        self.assertEqual(self.config_mngr.get_evaluation_name(), '')

        self.config_mngr._config[
            self.config_mngr.CONFIG_DEFAULT_SECTION
        ][
            self.config_mngr.CONFIG_EVALUATION_NAME
        ] = 'test_eval'

        self.assertEqual(
            self.config_mngr.get_evaluation_name(),
            'test_eval')

    def test_get_metadata_tier(self):
        self.assertEqual(
            self.config_mngr.get_metadata_tier(),
            MetadataTier.DEFAULT)
        self.assertEqual(self.config_mngr.get_metadata_tier().value, 'default')

        self.config_mngr._config[
            self.config_mngr.CONFIG_DEFAULT_SECTION
        ][
            self.config_mngr.CONFIG_METADATA_TIER
        ] = 'oracle'

        self.assertEqual(
            self.config_mngr.get_metadata_tier(),
            MetadataTier.ORACLE)
        self.assertEqual(
            self.config_mngr.get_metadata_tier().value,
            'oracle')

    def test_get_size(self):
        self.assertEqual(self.config_mngr.get_size(), 600)

        self.config_mngr._config[
            self.config_mngr.CONFIG_DEFAULT_SECTION
        ][
            self.config_mngr.CONFIG_SIZE
        ] = '800'

        self.assertEqual(self.config_mngr.get_size(), 800)

    def test_get_team(self):
        self.assertEqual(self.config_mngr.get_team(), '')

        self.config_mngr._config[
            self.config_mngr.CONFIG_DEFAULT_SECTION
        ][
            self.config_mngr.CONFIG_TEAM
        ] = 'team-name'

        self.assertEqual(
            self.config_mngr.get_team(),
            'team-name')

    def test_is_history_enabled(self):
        self.assertTrue(self.config_mngr.is_history_enabled())

        self.config_mngr._config[
            self.config_mngr.CONFIG_DEFAULT_SECTION
        ][
            self.config_mngr.CONFIG_HISTORY_ENABLED
        ] = 'false'

        self.config_mngr._config[
            self.config_mngr.CONFIG_DEFAULT_SECTION
        ][
            self.config_mngr.CONFIG_HISTORY_ENABLED
        ] = 'true'

        self.assertTrue(self.config_mngr.is_history_enabled())

    def test_is_noise_enabled(self):
        self.assertFalse(self.config_mngr.is_noise_enabled())

        self.config_mngr._config[
            self.config_mngr.CONFIG_DEFAULT_SECTION
        ][
            self.config_mngr.CONFIG_NOISE_ENABLED
        ] = 'true'

        self.assertTrue(self.config_mngr.is_noise_enabled())

        self.config_mngr._config[
            self.config_mngr.CONFIG_DEFAULT_SECTION
        ][
            self.config_mngr.CONFIG_NOISE_ENABLED
        ] = 'false'

        self.assertFalse(self.config_mngr.is_noise_enabled())

    def test_is_video_enabled(self):
        self.assertFalse(self.config_mngr.is_video_enabled())

        self.config_mngr._config[
            self.config_mngr.CONFIG_DEFAULT_SECTION
        ][
            self.config_mngr.CONFIG_VIDEO_ENABLED
        ] = 'true'

        self.assertTrue(self.config_mngr.is_video_enabled())

        self.config_mngr._config[
            self.config_mngr.CONFIG_DEFAULT_SECTION
        ][
            self.config_mngr.CONFIG_VIDEO_ENABLED
        ] = 'false'

        self.assertFalse(self.config_mngr.is_noise_enabled())

    def test_timeout(self):
        self.assertEqual(self.config_mngr.get_timeout(),
                         self.config_mngr.TIMEOUT_DEFAULT)

        self.config_mngr._config[
            self.config_mngr.CONFIG_DEFAULT_SECTION
        ][
            self.config_mngr.CONFIG_TIMEOUT
        ] = '50'

        self.assertEqual(self.config_mngr.get_timeout(), 50)


class TestSceneConfig(unittest.TestCase):
    def test_objects(self):
        object_config_list = [{
            'id': 'id_1',
            'type': 'type_1'
        }, {
            'id': 'id_2',
            'type': 'type_2',
            'actions': [{
                'stepBegin': 1,
                'stepEnd': 5,
                'id': 'animation_a',
                'isLoopAnimation': True,
            }, {
                'stepBegin': 25,
                'id': 'animation_b'
            }],
            'agentMovement': {
                "repeat": True,
                "stepBegin": 4,
                "sequence": [{
                    "animation": "TPM_walk",
                    "endPoint":
                    {
                        "x": 1,
                        "z": 1
                    }
                }, {
                    "animation": "TPM_run",
                    "endPoint":
                    {
                        "x": -0.7,
                        "z": -1.3
                    }
                }, {
                    "animation": "TPF_walk",
                    "endPoint":
                    {
                        "x": 1.2,
                        "z": -1.5
                    }
                }, {
                    "animation": "TPF_run",
                    "endPoint":
                    {
                        "x": -0.6,
                        "z": 1.2
                    }
                }]
            },
            'agentSettings': {
                'chest': 1,
                'chestMaterial': 2,
                'eyes': 3,
                'feet': 4,
                'feetMaterial': 5,
                'glasses': 6,
                'hair': 7,
                'hairMaterial': 8,
                'hatMaterial': 9,
                'hideHair': True,
                'isElder': True,
                'jacket': 10,
                'jacketMaterial': 11,
                'legs': 12,
                'legsMaterial': 13,
                'showBeard': True,
                'showGlasses': True,
                'showJacket': True,
                'showTie': True,
                'skin': 14,
                'tie': 15,
                'tieMaterial': 16
            },
            'associatedWithAgent': "agent_male",
            'centerOfMass': {
                'x': 0.01,
                'y': 0.02,
                'z': 0.03
            },
            'changeMaterials': [{
                'stepBegin': 10,
                'materials': ['material_3', 'material_4']
            }],
            'debug': {
                'key': 'value'
            },
            'forces': [{
                'impulse': True,
                'relative': True,
                'stepBegin': 11,
                'stepEnd': 12,
                'vector': {
                    'x': 0.04,
                    'y': 0.05,
                    'z': 0.06
                }
            }],
            'ghosts': [{
                'stepBegin': 13,
                'stepEnd': 14
            }],
            'hides': [{
                'stepBegin': 15
            }],
            'kinematic': True,
            'lidAttachment': {
                'stepBegin': 16,
                'lidAttachmentObjId': 'test_container'
            },
            'locationParent': 'parent_id',
            'mass': 12.34,
            'materials': ['material_1', 'material_2'],
            'maxAngularVelocity': 25,
            'moveable': True,
            'moves': [{
                'stepBegin': 16,
                'stepEnd': 17,
                'vector': {
                    'x': 0.07,
                    'y': 0.08,
                    'z': 0.09
                },
                'globalSpace': True
            }],
            'nullParent': {
                'position': {
                    'x': 0.11,
                    'y': 0.12,
                    'z': 0.13
                },
                'rotation': {
                    'x': 0.14,
                    'y': 0.15,
                    'z': 0.16
                },
                'scale': {
                    'x': 0.17,
                    'y': 0.18,
                    'z': 0.19
                }
            },
            'openable': True,
            'opened': True,
            'openClose': [{
                'step': 18,
                'open': True
            }],
            'physics': True,
            'physicsProperties': {
                'enable': True,
                'angularDrag': 1,
                'bounciness': 2,
                'drag': 3,
                'dynamicFriction': 4,
                'staticFriction': 5
            },
            'pickupable': True,
            'receptacle': True,
            'resetCenterOfMass': True,
            'resetCenterOfMassAtY': 0.321,
            'resizes': [{
                'stepBegin': 19,
                'stepEnd': 20,
                'size': {
                    'x': 0.21,
                    'y': 0.22,
                    'z': 0.23
                }
            }],
            'rotates': [{
                'stepBegin': 21,
                'stepEnd': 22,
                'vector': {
                    'x': 0.24,
                    'y': 0.25,
                    'z': 0.26
                }
            }],
            'salientMaterials': ['salient_1', 'salient_2'],
            'shows': [{
                'stepBegin': 23,
                'position': {
                    'x': 0.31,
                    'y': 0.32,
                    'z': 0.33
                },
                'rotation': {
                    'x': 0.34,
                    'y': 0.35,
                    'z': 0.36
                },
                'scale': {
                    'x': 0.37,
                    'y': 0.38,
                    'z': 0.39
                }
            }],
            'shrouds': [{
                'stepBegin': 24,
                'stepEnd': 25
            }],
            'states': [['state_1'], [], ['state_2'], ['state_1', 'state_3']],
            'structure': True,
            'teleports': [{
                'stepBegin': 26,
                'position': {
                    'x': 0.41,
                    'y': 0.42,
                    'z': 0.43
                }
            }],
            "triggeredBy": True,
            'togglePhysics': [{
                'stepBegin': 27
            }],
            'torques': [{
                'stepBegin': 28,
                'stepEnd': 29,
                'vector': {
                    'x': 0.44,
                    'y': 0.45,
                    'z': 0.46
                }
            }]
        }]
        object_list = [
            SceneObject(**object_config)
            for object_config in object_config_list
        ]

        toggle_lights_list = [StepBeginEndConfig(
            step_begin=4,
            step_end=5
        )]

        scene_config = SceneConfiguration(
            toggle_lights=toggle_lights_list,
            objects=object_list)
        self.assertEqual(len(scene_config.objects), 2)

        self.assertEqual(scene_config.toggle_lights, [StepBeginEndConfig(
            step_begin=4,
            step_end=5
        )])

        object_1 = scene_config.objects[0]
        object_2 = scene_config.objects[1]

        self.assertEqual(object_1.id, 'id_1')
        self.assertEqual(object_1.type, 'type_1')
        self.assertIsNone(object_1.actions)
        self.assertIsNone(object_1.agent_movement)
        self.assertIsNone(object_1.agent_settings)
        self.assertIsNone(object_1.center_of_mass)
        self.assertIsNone(object_1.change_materials)
        self.assertIsNone(object_1.debug)
        self.assertIsNone(object_1.forces)
        self.assertIsNone(object_1.ghosts)
        self.assertIsNone(object_1.hides)
        self.assertIsNone(object_1.kinematic)
        self.assertIsNone(object_1.lid_attachment)
        self.assertIsNone(object_1.location_parent)
        self.assertIsNone(object_1.mass)
        self.assertIsNone(object_1.materials)
        self.assertIsNone(object_1.max_angular_velocity)
        self.assertIsNone(object_1.moveable)
        self.assertIsNone(object_1.moves)
        self.assertIsNone(object_1.null_parent)
        self.assertIsNone(object_1.openable)
        self.assertIsNone(object_1.opened)
        self.assertIsNone(object_1.open_close)
        self.assertIsNone(object_1.physics)
        self.assertIsNone(object_1.physics_properties)
        self.assertIsNone(object_1.pickupable)
        self.assertIsNone(object_1.receptacle)
        self.assertIsNone(object_1.reset_center_of_mass)
        self.assertIsNone(object_1.reset_center_of_mass_at_y)
        self.assertIsNone(object_1.resizes)
        self.assertIsNone(object_1.rotates)
        self.assertIsNone(object_1.salient_materials)
        self.assertIsNone(object_1.shows)
        self.assertIsNone(object_1.shrouds)
        self.assertIsNone(object_1.states)
        self.assertIsNone(object_1.structure)
        self.assertIsNone(object_1.triggered_by)
        self.assertIsNone(object_1.teleports)
        self.assertIsNone(object_1.toggle_physics)
        self.assertIsNone(object_1.torques)

        self.assertEqual(object_2.id, 'id_2')
        self.assertEqual(object_2.type, 'type_2')
        self.assertEqual(object_2.actions, [
            ActionConfig(
                step_begin=1,
                id='animation_a',
                step_end=5,
                is_loop_animation=True),
            ActionConfig(
                step_begin=25,
                id='animation_b',
                is_loop_animation=False)
        ])
        self.assertEqual(object_2.agent_movement, AgentMovementConfig(
            repeat=True,
            step_begin=4,
            sequence=[
                SequenceConfig(
                    animation="TPM_walk",
                    end_point=Vector3d(x=1, y=0, z=1)),
                SequenceConfig(
                    animation="TPM_run",
                    end_point=Vector3d(x=-0.7, y=0, z=-1.3)),
                SequenceConfig(
                    animation="TPF_walk",
                    end_point=Vector3d(x=1.2, y=0, z=-1.5)),
                SequenceConfig(
                    animation="TPF_run",
                    end_point=Vector3d(x=-0.6, y=0, z=1.2))
            ]
        ))
        self.assertEqual(object_2.agent_settings, AgentSettings(
            chest=1,
            chest_material=2,
            eyes=3,
            feet=4,
            feet_material=5,
            glasses=6,
            hair=7,
            hair_material=8,
            hat_material=9,
            hide_hair=True,
            is_elder=True,
            jacket=10,
            jacket_material=11,
            legs=12,
            legs_material=13,
            show_beard=True,
            show_glasses=True,
            show_jacket=True,
            show_tie=True,
            skin=14,
            tie=15,
            tie_material=16
        ))
        self.assertEqual(object_2.associated_with_agent, 'agent_male')
        self.assertEqual(
            object_2.center_of_mass, Vector3d(
                x=0.01, y=0.02, z=0.03))
        self.assertEqual(object_2.change_materials, [ChangeMaterialConfig(
            step_begin=10,
            materials=['material_3', 'material_4']
        )])
        self.assertEqual(object_2.debug, {'key': 'value'})
        self.assertEqual(object_2.forces, [ForceConfig(
            impulse=True,
            relative=True,
            step_begin=11,
            step_end=12,
            vector=Vector3d(x=0.04, y=0.05, z=0.06)
        )])
        self.assertEqual(object_2.ghosts, [StepBeginEndConfig(
            step_begin=13,
            step_end=14
        )])
        self.assertEqual(object_2.hides, [SingleStepConfig(step_begin=15)])
        self.assertTrue(object_2.kinematic)
        self.assertEqual(object_2.lid_attachment, LidConfig(
            step_begin=16,
            lid_attachment_obj_id='test_container'
        ))
        self.assertEqual(object_2.location_parent, 'parent_id')
        self.assertEqual(object_2.mass, 12.34)
        self.assertListEqual(object_2.materials, ['material_1', 'material_2'])
        self.assertEqual(object_2.max_angular_velocity, 25)
        self.assertTrue(object_2.moveable)
        self.assertEqual(object_2.moves, [MoveConfig(
            step_begin=16,
            step_end=17,
            vector=Vector3d(x=0.07, y=0.08, z=0.09),
            global_space=True
        )])
        self.assertEqual(object_2.null_parent, TransformConfig(
            position=Vector3d(x=0.11, y=0.12, z=0.13),
            rotation=Vector3d(x=0.14, y=0.15, z=0.16),
            scale=Vector3d(x=0.17, y=0.18, z=0.19)
        ))
        self.assertTrue(object_2.openable)
        self.assertTrue(object_2.opened)
        self.assertEqual(
            object_2.open_close, [
                OpenCloseConfig(
                    open=True, step=18)])
        self.assertTrue(object_2.physics)
        self.assertEqual(object_2.physics_properties, PhysicsConfig(
            enable=True,
            angular_drag=1,
            bounciness=2,
            drag=3,
            dynamic_friction=4,
            static_friction=5
        ))
        self.assertTrue(object_2.pickupable)
        self.assertTrue(object_2.receptacle)
        self.assertTrue(object_2.reset_center_of_mass)
        self.assertEqual(object_2.reset_center_of_mass_at_y, 0.321)
        self.assertEqual(object_2.resizes, [SizeConfig(
            step_begin=19,
            step_end=20,
            size=Vector3d(x=0.21, y=0.22, z=0.23)
        )])
        self.assertEqual(object_2.rotates, [MoveConfig(
            step_begin=21,
            step_end=22,
            vector=Vector3d(x=0.24, y=0.25, z=0.26)
        )])
        self.assertListEqual(
            object_2.salient_materials, [
                'salient_1', 'salient_2'])
        self.assertEqual(object_2.shows, [ShowConfig(
            step_begin=23,
            position=Vector3d(x=0.31, y=0.32, z=0.33),
            rotation=Vector3d(x=0.34, y=0.35, z=0.36),
            scale=Vector3d(x=0.37, y=0.38, z=0.39)
        )])
        self.assertEqual(object_2.shrouds, [StepBeginEndConfig(
            step_begin=24,
            step_end=25
        )])
        self.assertListEqual(object_2.states, [
            ['state_1'], [], ['state_2'], ['state_1', 'state_3']
        ])
        self.assertTrue(object_2.structure)
        self.assertEqual(object_2.teleports, [TeleportConfig(
            step_begin=26,
            position=Vector3d(x=0.41, y=0.42, z=0.43)
        )])
        self.assertEqual(
            object_2.triggered_by, True)
        self.assertEqual(
            object_2.toggle_physics, [
                SingleStepConfig(
                    step_begin=27)])
        self.assertEqual(object_2.torques, [ForceConfig(
            step_begin=28,
            step_end=29,
            vector=Vector3d(x=0.44, y=0.45, z=0.46)
        )])

    def test_is_passive_scene(self):
        goal_1 = Goal(category='retrieval')
        goal_2 = Goal(category='imitation')
        goal_3 = Goal(category='passive')
        goal_4 = Goal(category='intuitive physics')
        goal_5 = Goal(category='agents')

        config_1 = SceneConfiguration(goal=goal_1)
        config_2 = SceneConfiguration(goal=goal_2)
        config_3 = SceneConfiguration(goal=goal_3)
        config_4 = SceneConfiguration(goal=goal_4)
        config_5 = SceneConfiguration(goal=goal_5)

        assert not config_1.is_passive_scene()
        assert not config_2.is_passive_scene()
        assert config_3.is_passive_scene()
        assert config_4.is_passive_scene()
        assert config_5.is_passive_scene()

        config_6 = SceneConfiguration()
        config_7 = SceneConfiguration(intuitive_physics=True)
        config_8 = SceneConfiguration(isometric=True)

        assert not config_6.is_passive_scene()
        assert config_7.is_passive_scene()
        assert config_8.is_passive_scene()

    def test_retrieve_goal_with_config_metadata(self):
        # self.controller.set_metadata_tier('oracle')
        goal = {
            'metadata': {
                'target': {'image': [0]},
                'target_1': {'image': [1]},
                'target_2': {'image': [2]},
                'targets': [{'image': [3]}]
            }
        }

        goal = Goal(**goal)
        scene_config = SceneConfiguration(name="test", version=1, goal=goal)
        actual = scene_config.retrieve_goal()
        self.assertEqual(actual.metadata, {
            'target': {'image': [0]},
            'target_1': {'image': [1]},
            'target_2': {'image': [2]},
            'targets': [{'image': [3]}]
        })

    def test_retrieve_goal(self):
        goal = Goal(**{})
        scene_config = SceneConfiguration(name="test", version=1, goal=goal)
        goal_1 = scene_config.retrieve_goal()
        self.assertEqual(goal_1.action_list, None)
        self.assertEqual(goal_1.category, '')
        self.assertEqual(goal_1.description, '')
        self.assertEqual(goal_1.habituation_total, 0)
        self.assertEqual(goal_1.last_step, None)
        self.assertEqual(goal_1.metadata, {})

        goal = {
            "goal": {
            }
        }
        goal = Goal(**{})
        scene_config = SceneConfiguration(
            name="test", version=1, goal=goal)
        goal_2 = scene_config.retrieve_goal()
        self.assertEqual(goal_2.action_list, None)
        self.assertEqual(goal_2.category, '')
        self.assertEqual(goal_2.description, '')
        self.assertEqual(goal_2.habituation_total, 0)
        self.assertEqual(goal_2.last_step, None)
        self.assertEqual(goal_2.metadata, {})

        goal = {
            "action_list": [
                [(mcs.Action.MOVE_AHEAD.value, {"amount": 0.1})],
                [],
                [(mcs.Action.PASS.value, {}),
                 (mcs.Action.ROTATE_LEFT.value, {}),
                 (mcs.Action.ROTATE_RIGHT.value, {})]
            ],
            "category": "test category",
            "description": "test description",
            "habituation_total": 5,
            "last_step": 10,
            "metadata": {
                "key": "value"
            }
        }
        goal = Goal(**goal)
        scene_config = SceneConfiguration(
            name="test", version=1, goal=goal)
        goal_3 = scene_config.retrieve_goal()

        self.assertEqual(goal_3.action_list, [
            [(mcs.Action.MOVE_AHEAD.value, {"amount": 0.1})],
            [],
            [(mcs.Action.PASS.value, {}), (mcs.Action.ROTATE_LEFT.value, {}),
             (mcs.Action.ROTATE_RIGHT.value, {})]
        ])
        self.assertEqual(goal_3.category, "test category")
        self.assertEqual(goal_3.description, "test description")
        self.assertEqual(goal_3.habituation_total, 5)
        self.assertEqual(goal_3.last_step, 10)
        self.assertEqual(goal_3.metadata, {
            "category": "test category",
            "key": "value"
        })

    def test_update_goal_target_image(self):
        goal = {'metadata': {
            'target': {'image': [0]},
            'target_1': {'image': [1]},
            'target_2': {'image': [2]},
            'targets': [{'image': [3]}]
        }}
        goal = Goal(**goal)
        scene_config = SceneConfiguration(name="test", version=1, goal=goal)
        actual = scene_config.update_goal_target_image(scene_config.goal)
        self.assertEqual(actual.metadata, {
            'target': {'image': [0]},
            'target_1': {'image': [1]},
            'target_2': {'image': [2]},
            'targets': [{'image': [3]}]
        })

    def test_update_goal_target_image_img_as_str(self):
        goal = {'metadata': {
            'target': {'image': "[0]"},
            'target_1': {'image': "[1]"},
            'target_2': {'image': "[2]"},
            'targets': [{'image': "[3]"}]
        }}
        goal = Goal(**goal)
        scene_config = SceneConfiguration(name="test", version=1, goal=goal)
        actual = scene_config.update_goal_target_image(scene_config.goal)
        self.assertEqual(actual.metadata, {
            'target': {'image': [0]},
            'target_1': {'image': [1]},
            'target_2': {'image': [2]},
            'targets': [{'image': [3]}]
        })

    def test_update_goal_target_image_oracle(self):
        goal = {'metadata': {
            'target': {'image': [0]},
            'target_1': {'image': [1]},
            'target_2': {'image': [2]},
            'targets': [{'image': [3]}]
        }}
        goal = Goal(**goal)
        scene_config = SceneConfiguration(name="test", version=1, goal=goal)
        actual = scene_config.update_goal_target_image(scene_config.goal)
        self.assertEqual(actual.metadata, {
            'target': {'image': [0]},
            'target_1': {'image': [1]},
            'target_2': {'image': [2]},
            'targets': [{'image': [3]}]
        })

    def test_update_goal_target_image_oracle_img_as_str(self):
        goal = {'metadata': {
            'target': {'image': "[0]"},
            'target_1': {'image': "[1]"},
            'target_2': {'image': "[2]"},
            'targets': [{'image': "[3]"}]
        }}
        goal = Goal(**goal)
        scene_config = SceneConfiguration(name="test", version=1, goal=goal)
        actual = scene_config.update_goal_target_image(scene_config.goal)
        self.assertEqual(actual.metadata, {
            'target': {'image': [0]},
            'target_1': {'image': [1]},
            'target_2': {'image': [2]},
            'targets': [{'image': [3]}]
        })


class TestSceneConfiguration(unittest.TestCase):

    def setUp(self):
        self.scheme_config = SceneConfiguration()

    @unittest.skip
    def test_remove_none(self):
        actual = self.scheme_config.remove_none({})
        self.assertEqual({}, actual)

        actual = self.scheme_config.remove_none({"test": None})
        self.assertEqual({}, actual)

        actual = self.scheme_config.remove_none({"test": 1})
        self.assertEqual({"test": 1}, actual)

        actual = self.scheme_config.remove_none({"test": 1, "none": None})
        self.assertEqual({"test": 1}, actual)

        actual = self.scheme_config.remove_none({"test1": {"test2": 1}})
        self.assertEqual({"test1": {"test2": 1}}, actual)

        actual = self.scheme_config.remove_none(
            {"test1": {"test2": 1, "none": None}})
        self.assertEqual({"test1": {"test2": 1}}, actual)

        actual = self.scheme_config.remove_none(
            {"test1": {"test2": 1}, "none": None})
        self.assertEqual({"test1": {"test2": 1}}, actual)

        actual = self.scheme_config.remove_none(
            {"test1": {"test2": 1, "none": None}, "none": None})
        self.assertEqual({"test1": {"test2": 1}}, actual)

        actual = self.scheme_config.remove_none({"test1": [{"test2": None}]})
        self.assertEqual({"test1": [{}]}, actual)

        actual = self.scheme_config.remove_none(
            {"test1": [{"test2": None, "test3": "test"}]})
        self.assertEqual({"test1": [{"test3": "test"}]}, actual)

        actual = self.scheme_config.remove_none(
            {"test1": [{"test2": None}], "test3": False})
        self.assertEqual({"test1": [{}], "test3": False}, actual)


if __name__ == '__main__':
    unittest.main()



from typing import Dict, List

from .stringifier import Stringifier


class SceneHistory(object):
    def __init__(
        self,
        step=-1,
        action=None,  # MCS-1181 Enum vs string
        args=None,
        params=None,
        # Use classification and confidence rather than rating and score to be
        # compatible with old history files.
        classification: str = None,
        confidence: float = None,
        violations_xy_list: List[Dict[str, float]] = None,
        internal_state: object = None,
        delta_time_millis=0,
        output=None,
        target_visible=False,
        target_is_visible_at_start=None
    ):
        self.step = step
        self.action = action
        self.args = args
        self.params = params
        self.classification = classification
        self.confidence = confidence
        self.violations_xy_list = violations_xy_list
        self.internal_state = internal_state
        self.delta_time_millis = delta_time_millis
        self.output = output
        self.target_visible = target_visible
        self.target_is_visible_at_start = target_is_visible_at_start

    def __str__(self):
        return Stringifier.class_to_str(self)

    # Allows converting the class to a dictionary, along with allowing
    #   certain fields to be left out of output file
    def __iter__(self):
        yield 'step', self.step
        yield 'action', self.action
        yield 'args', self.args
        yield 'params', self.params
        yield 'classification', self.classification
        yield 'confidence', self.confidence
        yield 'violations_xy_list', self.violations_xy_list
        yield 'internal_state', self.internal_state
        yield 'output', dict(self.output) if(
            self.output) is not None else self.output
        yield 'delta_time_millis', self.delta_time_millis
        yield 'target_visible', self.target_visible
        if self.target_is_visible_at_start is not None:
            yield 'target_is_visible_at_start', (
                self.target_is_visible_at_start)



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()


def get_version():
    version_file = "machine_common_sense/_version.py"
    with open(version_file) as f:
        exec(compile(f.read(), version_file, "exec"))
    return locals()["__version__"]


MCS_VERSION = get_version()


setuptools.setup(
    name='machine_common_sense',
    version=MCS_VERSION,
    maintainer='Next Century, a wholly owned subsidiary of CACI',
    maintainer_email='mcs-ta2@machinecommonsense.com',
    url='https://github.com/NextCenturyCorporation/MCS/',
    description=('Machine Common Sense Python API'
                 ' to Unity 3D Simulation Environment'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Games/Entertainment :: Simulation',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Multimedia :: Graphics :: 3D Rendering',
        'Typing :: Typed'
    ],
    license='Apache-2',
    python_requires=">=3.7,<3.10",
    packages=setuptools.find_packages(),
    package_data={'': ['*.ini'], },
    install_requires=[
        'shapely>=1.7.0',
        'colour>=0.1.5',
        'opencv-python>=4.0',
        'matplotlib>=3.3',
        'msgpack>=1.0.0',
        'numpyencoder>=0.3.0',
        'ai2thor==2.5.0',
        'scikit-image>=0.17.1',
        'typeguard>=2.11.1',
        'pydantic>=1.9.0'
    ],
    entry_points={
        'console_scripts': [
            ('run_in_human_input_mode='
             'machine_common_sense.scripts.run_human_input:main'),
            ('run_scene_timer='
             'machine_common_sense.scripts.run_scene_timer:main'),
            ('cache_addressables='
             'machine_common_sense.scripts.cache_addressables:main')
        ]
    }
)



from runner_script import SingleFileRunnerScript


def action_callback(scene_data, step_metadata, runner_script):
    last_step = 60
    if (
        'goal' in scene_data.keys(
        ) and 'last_step' in scene_data['goal'].keys()
    ):
        last_step = scene_data['goal']['last_step']
    if step_metadata.step_number <= last_step:
        action, params = (
            step_metadata.action_list[len(step_metadata.action_list) - 1]
            if len(step_metadata.action_list) else (None, None)
        )
        return action, params
    return None, None


def main():
    SingleFileRunnerScript('Last Action', action_callback)


if __name__ == "__main__":
    main()



import os

import ai2thor.server
import numpy

from machine_common_sense.config_manager import ConfigManager
from machine_common_sense.controller import Controller
from machine_common_sense.controller_output_handler import \
    ControllerOutputHandler

MOCK_VARIABLES = {
    'event_count': 5,
    'frame': numpy.array([[0]], dtype=numpy.uint8),
    'depth_frame': numpy.array([[[0, 0, 0]]], dtype=numpy.uint8),
    'instance_segmentation_frame': numpy.array([[0]], dtype=numpy.uint8),
    'metadata': {
        'agent': {
            'cameraHorizon': 0.0,
            'position': {
                'x': 0.0,
                'y': 0.0,
                'z': 0.0
            },
            'rotation': {
                'y': 0.0
            }
        },
        'lastActionStatus': 'SUCCESSFUL',
        'objects': [],
        'screenHeight': 400,
        'screenWidth': 600,
        'structuralObjects': []
    }
}


class MockController():
    '''Mock of the Controller class from the AI2-THOR library.'''

    def __init__(self):
        self.__last_step_data = None
        self.__last_metadata = MOCK_VARIABLES['metadata'].copy()
        self._subscribers = []

    def step(self, data):
        self.__last_step_data = data
        metadata = self.__last_metadata

        event = ai2thor.server.Event(metadata)
        event.frame = MOCK_VARIABLES['frame'].copy()
        event.depth_frame = MOCK_VARIABLES['depth_frame'].copy()
        event.instance_segmentation_frame = MOCK_VARIABLES[
            'instance_segmentation_frame'
        ].copy()
        return ai2thor.server.MultiAgentEvent(
            0, [event for _ in range(MOCK_VARIABLES['event_count'])]
        )

    def get_last_step_data(self):
        return self.__last_step_data

    def subscribe(self, subscriber):
        if subscriber not in self._subscribers:
            self._subscribers.append(subscriber)


class MockControllerAI2THOR(Controller):
    '''Mock of the ControllerAI2THOR class from the MCS library.'''

    def __init__(self):
        # Do NOT call superclass __init__ function

        # Need to clear any previously set environment variables
        # so that they don't affect test cases
        check_config_path = os.getenv(
            ConfigManager.CONFIG_FILE_ENV_VAR, None)

        if(check_config_path is not None):
            os.environ.pop(ConfigManager.CONFIG_FILE_ENV_VAR)

        self._subscribers = []

        self._failure_handler_registered = True  # atexit not needed for tests
        self._end_scene_called = False
        self._controller = MockController()
        self._config = ConfigManager(config_file_or_dict={})
        self._config._config[
            ConfigManager.CONFIG_DEFAULT_SECTION
        ] = {}
        self._output_handler = ControllerOutputHandler(self._config)
        self._on_init()
        self._set_config(self._config)

    def get_last_step_data(self):
        return self._controller.get_last_step_data()

    def set_goal(self, goal):
        self._goal = goal

    def set_metadata_tier(self, mode):
        if not self._config:
            self._config = ConfigManager()
        self._config.set_metadata_tier(mode)



from runner_script import SingleFileRunnerScript

import machine_common_sense as mcs

action_list_from_file = []


class ActionFileRunnerScript(SingleFileRunnerScript):
    def _append_subclass_args_to_parser(self, parser):
        parser = super()._append_subclass_args_to_parser(parser)
        parser.add_argument(
            'action_filename',
            help='Filename of MCS actions to run'
        )
        return parser


def action_callback(scene_data, step_metadata, runner_script):
    if not action_list_from_file:
        with open(runner_script.args.action_filename, 'r') as action_file:
            for line in action_file:
                action_list_from_file.append(line.strip())

    if len(step_metadata.action_list) == 1:
        return step_metadata.action_list[0]

    if len(action_list_from_file) <= step_metadata.step_number:
        return None, None

    return mcs.Action.input_to_action_and_params(
        action_list_from_file[step_metadata.step_number]
    )


def main():
    ActionFileRunnerScript('Action File', action_callback)


if __name__ == "__main__":
    main()



import copy
import logging
from dataclasses import dataclass
from typing import Tuple

import numpy as np
import PIL
from ai2thor.server import Event

from .config_manager import ConfigManager, MetadataTier, SceneConfiguration
from .controller import DEFAULT_MOVE
from .material import Material
from .object_metadata import ObjectMetadata
from .return_status import ReturnStatus
from .reward import Reward
from .step_metadata import StepMetadata

logger = logging.getLogger(__name__)


@dataclass
class SceneEvent():
    '''Wraps step output from AI2thor'''
    _config: ConfigManager
    _scene_config: SceneConfiguration
    _raw_output: Event
    _step_number: int

    def __post_init__(self):
        self.image_list = []
        self.depth_map_list = []
        self.object_mask_list = []

        for event in self.events:
            if hasattr(event, 'frame'):
                scene_image = PIL.Image.fromarray(event.frame)
                self.image_list.append(scene_image)

                if self._config.is_depth_maps_enabled():
                    unity_depth_array = event.depth_frame.astype(np.float32)
                    # Convert to a 2D array (screen length X width)
                    depth_float_array = np.squeeze(unity_depth_array)
                    # Convert from (0.0, 1.0) to (0, max distance)
                    depth_float_array = (
                        depth_float_array * self.clipping_plane_far
                    )
                    self.depth_map_list.append(np.array(depth_float_array))

                if self._config.is_object_masks_enabled():
                    object_mask = PIL.Image.fromarray(
                        event.instance_segmentation_frame)
                    self.object_mask_list.append(object_mask)

    @property
    def objects(self) -> list:
        return self._raw_output.metadata.get('objects', None)

    @property
    def agent(self) -> dict:
        return self._raw_output.metadata.get('agent', None)

    @property
    def head_tilt(self) -> float:
        return self._raw_output.metadata['agent']['cameraHorizon']

    @property
    def rotation(self) -> float:
        return self._raw_output.metadata['agent']['rotation']['y']

    @property
    def camera_field_of_view(self):
        return self._raw_output.metadata.get('fov', 0.0)

    @property
    def clipping_plane(self):
        return (self.clipping_plane_near, self.clipping_plane_far)

    @property
    def clipping_plane_near(self):
        return self._raw_output.metadata.get('clippingPlaneNear', 0.0)

    @property
    def clipping_plane_far(self):
        return self._raw_output.metadata.get('clippingPlaneFar', 0.0)

    @property
    def haptic_feedback(self):
        return self._raw_output.metadata.get('hapticFeedback')

    @property
    def resolved_object(self):
        return self._raw_output.metadata.get('resolvedObject', '')

    @property
    def resolved_receptacle(self):
        return self._raw_output.metadata.get('resolvedReceptacle', '')

    @property
    def steps_on_lava(self):
        return self._raw_output.metadata.get('stepsOnLava')

    @property
    def triggered_by_sequence_incorrect(self):
        return self._raw_output.metadata.get('triggeredBySequenceIncorrect')

    @property
    def performer_radius(self):
        return self._raw_output.metadata.get('performerRadius')

    @property
    def performer_reach(self):
        return self._raw_output.metadata.get('performerReach')

    @property
    def physics_frames_per_second(self) -> float:
        return self._raw_output.metadata.get('physicsFramesPerSecond')

    @property
    def room_dimensions(self) -> float:
        return self._raw_output.metadata.get('roomDimensions')

    @property
    def events(self):
        return self._raw_output.events

    @property
    def camera_height(self):
        return self._raw_output.metadata.get(
            'cameraPosition', {}).get('y', 0.0)

    @property
    def position(self) -> dict:
        return self._raw_output.metadata['agent']['position']

    def _get_objects(self, key: str):
        # Return object list for all tier levels, the restrict output function
        # will then strip out the necessary metadata
        metadata_tier = self._config.get_metadata_tier()
        show_all = metadata_tier != MetadataTier.DEFAULT
        # if no config specified, return visible objects (for now)
        return sorted(
            [
                self.retrieve_object_output(
                    object_metadata, self.object_colors
                )
                for object_metadata in self._raw_output.metadata[key]
                if show_all or object_metadata['visibleInCamera'] or
                object_metadata['isPickedUp']
            ],
            key=lambda x: x.uuid
        )

    @property
    def object_list(self):
        return self._get_objects('objects')

    @property
    def structural_object_list(self):
        return self._get_objects('structuralObjects')

    @property
    def return_status(self):
        return_status = ReturnStatus.UNDEFINED.name

        try:
            if self._raw_output.metadata['lastActionStatus']:
                return_status = ReturnStatus[
                    self._raw_output.metadata['lastActionStatus']
                ].name
        except KeyError:
            logger.error(
                f"Return status "
                f"{self._raw_output.metadata['lastActionStatus']}"
                " is not currently supported.")
        finally:
            return return_status

    @property
    def object_colors(self):
        # Use the color map for the final event (though they should all be the
        # same anyway).
        event = self._raw_output.events[len(
            self._raw_output.events) - 1]
        return event.object_id_to_color

    @property
    def segmentation_colors(self):
        return [{
            'objectId': item['name'],
            'r': item['color'][0],
            'g': item['color'][1],
            'b': item['color'][2]
        } for item in (self._raw_output.metadata.get('colors') or [])]

    def retrieve_object_output(
            self, object_metadata, object_id_to_color):
        material_list = (
            list(
                filter(
                    Material.verify_material_enum_string,
                    [
                        material.upper()
                        for material in object_metadata['salientMaterials']
                    ],
                )
            )
            if object_metadata['salientMaterials'] is not None
            else []
        )

        rgb = (
            object_id_to_color[object_metadata['objectId']]
            if object_metadata['objectId'] in object_id_to_color
            else [None, None, None]
        )

        bounds = (
            object_metadata['objectBounds']
            if 'objectBounds' in object_metadata and
            object_metadata['objectBounds'] is not None
            else {}
        )

        return ObjectMetadata(
            uuid=object_metadata['objectId'],
            dimensions=(
                bounds['objectBoundsCorners']
                if 'objectBoundsCorners' in bounds
                else None
            ),
            direction=object_metadata['direction'],
            distance=(
                object_metadata['distanceXZ'] / DEFAULT_MOVE
            ),  # DEPRECATED
            distance_in_steps=(
                object_metadata['distanceXZ'] / DEFAULT_MOVE
            ),
            distance_in_world=(object_metadata['distance']),
            held=object_metadata['isPickedUp'],
            mass=object_metadata['mass'],
            material_list=material_list,
            position=object_metadata['position'],
            rotation=object_metadata['rotation'],
            segment_color={'r': rgb[0], 'g': rgb[1], 'b': rgb[2]},
            shape=object_metadata['shape'],
            state_list=self._scene_config.retrieve_object_states(
                object_metadata['objectId'],
                self._step_number
            ),
            texture_color_list=object_metadata['colorsFromMaterials'],
            visible=(
                object_metadata['visibleInCamera'] or
                object_metadata['isPickedUp']
            ),
            is_open=object_metadata['isOpen'],
            openable=object_metadata['openable'],
            locked=object_metadata['locked'],
            associated_with_agent=object_metadata['associatedWithAgent'],
            simulation_agent_held_object=object_metadata[
                'simulationAgentHeldObject'],
            simulation_agent_is_holding_held_object=object_metadata[
                'simulationAgentIsHoldingHeldObject']
        )


class ControllerOutputHandler():
    '''
    Attempts to handle converting ai2thor output in to MCS output.
    This class will be refactored again in MCS-665
    '''

    def __init__(self, config: ConfigManager):
        # do nothing
        self._config = config

    def set_scene_config(self, scene_config):
        self._scene_config = scene_config

    def handle_output(self, raw_output, goal, step_number, habituation_trial):
        self._scene_event = SceneEvent(
            self._config, self._scene_config, raw_output, step_number)
        self._step_number = step_number
        unrestricted = self._get_step_metadata(
            goal, habituation_trial, False)
        restricted = self._get_step_metadata(
            goal, habituation_trial, True)
        return (unrestricted, restricted)

    def _get_step_metadata(self, goal, habituation_trial,
                           restricted=True) -> StepMetadata:
        (restrict_depth_map, restrict_object_mask_list, restrict_non_oracle) =\
            self.get_restrictions(restricted, self._config.get_metadata_tier())
        step_output = StepMetadata(
            action_list=goal.retrieve_action_list_at_step(
                self._step_number,
                self._scene_event.steps_on_lava,
                self._scene_event.triggered_by_sequence_incorrect,
                self._scene_config.is_passive_scene()
            ),
            camera_aspect_ratio=self._config.get_screen_size(),
            camera_clipping_planes=self._scene_event.clipping_plane,
            camera_field_of_view=self._scene_event.camera_field_of_view,
            camera_height=self._scene_event.camera_height,
            depth_map_list=[] if restrict_depth_map else (
                self._scene_event.depth_map_list),
            goal=goal,
            habituation_trial=(
                habituation_trial
                if goal.habituation_total >= habituation_trial
                else None
            ),
            haptic_feedback=self._scene_event.haptic_feedback,
            head_tilt=self._scene_event.head_tilt,
            holes=None if restrict_non_oracle else [
                (hole.x, hole.z) for hole in self._scene_config.holes
            ],
            image_list=self._scene_event.image_list,
            lava=(
                None if restrict_non_oracle else
                self._scene_config.retrieve_lava()
            ),
            object_list=(
                [] if restrict_non_oracle else self._scene_event.object_list),
            object_mask_list=([] if restrict_object_mask_list else
                              self._scene_event.object_mask_list),
            position=(
                None if restrict_non_oracle else self._scene_event.position),
            performer_radius=self._scene_event.performer_radius,
            performer_reach=self._scene_event.performer_reach,
            return_status=self._scene_event.return_status,
            reward=Reward.calculate_reward(
                goal, self._scene_event.objects, self._scene_event.agent,
                self._step_number, self._scene_event.performer_reach,
                self._scene_event.steps_on_lava,
                self._config.get_lava_penalty(),
                self._config.get_step_penalty(),
                self._config.get_goal_reward()),
            resolved_object=(
                None if restrict_non_oracle
                else self._scene_event.resolved_object),
            resolved_receptacle=(
                None if restrict_non_oracle
                else self._scene_event.resolved_receptacle),
            room_dimensions=(
                None if restrict_non_oracle
                else self._scene_event.room_dimensions),
            rotation=(
                None if restrict_non_oracle else self._scene_event.rotation),
            segmentation_colors=(
                None if restrict_non_oracle
                else self._scene_event.segmentation_colors
            ),
            step_number=self._step_number,
            steps_on_lava=self._scene_event.steps_on_lava,
            triggered_by_sequence_incorrect=(
                self._scene_event.triggered_by_sequence_incorrect),
            physics_frames_per_second=(
                self._scene_event.physics_frames_per_second),
            structural_object_list=([] if restrict_non_oracle else
                                    self._scene_event.structural_object_list)
        )

        if (restrict_non_oracle):
            self.filter_step_output(step_output)

        return step_output

    def get_restrictions(self, restricted, metadata_tier) -> Tuple:
        restrict_depth_map = (
            restricted and
            metadata_tier == MetadataTier.NONE
        )

        restrict_object_mask_list = restricted and metadata_tier in [
            MetadataTier.NONE,
            MetadataTier.LEVEL_1,
        ]

        restrict_non_oracle = restricted and metadata_tier in [
            MetadataTier.NONE,
            MetadataTier.LEVEL_1,
            MetadataTier.LEVEL_2,
        ]
        return (restrict_depth_map, restrict_object_mask_list,
                restrict_non_oracle)

    def filter_step_output(self, step_output):
        # This is here to retain the exact outputs as before
        # If None is passed into the constructor, it is turned into
        # {}, but restricted mode turns it to None.
        step_output.position = None
        step_output.holes = None
        step_output.lava = None
        step_output.room_dimensions = None

        # Copy the goal object to avoid deleting data from the original object
        step_output.goal = copy.deepcopy(step_output.goal)

        step_output.goal.triggered_by_target_sequence = None
        metadata = step_output.goal.metadata or {}
        # Different goal categories may use different property names
        target_names = ['target', 'targets', 'target_1', 'target_2']
        for target_name in target_names:
            target = metadata.get(target_name)
            # Some properties may be dicts, and some may be lists of dicts
            if isinstance(target, list):
                # Clear the whole array, to avoid revealing the total count
                metadata[target_name] = []
            elif target:
                # Pretty sure the target always has an id, but just in case...
                if 'id' in target:
                    del target['id']
                # Backwards compatibility: target used to have image data
                if 'image' in target:
                    del target['image']
                if 'image_name' in target:
                    del target['image_name']



import unittest

import machine_common_sense as mcs


class MyEmptyclass:

    def __init__(self):
        pass


class MySubclass:

    def __init__(self):
        self.my_integer = 7
        self.my_string = "h"
        self.my_list = [8, "i"]
        self.my_dict = {
            "my_integer": 9,
            "my_string": "j",
        }

    def __str__(self):
        return mcs.Stringifier.class_to_str(self)


class MyClass:

    def __init__(self):
        self.my_boolean = True
        self.my_float = 1.234
        self.my_integer = 0
        self.my_string = "a"
        self.my_list = [1, "b", {
            "my_integer": 2,
            "my_string": "c",
            "my_list": [3, "d"]
        }]
        self.my_dict = {
            "my_integer": 4,
            "my_string": "e",
            "my_list": [5, "f"],
            "my_dict": {
                "my_integer": 6,
                "my_string": "g",
            }
        }
        self.my_list_empty = []
        self.my_dict_empty = {}
        self.my_subclass = MySubclass()
        self.__my_private = "z"

    def my_function():
        pass


class TestStringifier(unittest.TestCase):

    def test_class_to_str_with_class(self):
        self.maxDiff = 10000
        expected = "{\n    \"my_boolean\": true,\n    \"my_float\": 1.234,\n    \"my_integer\": 0,\n    \"my_string\": \"a\",\n    \"my_list\": [\n        1,\n        \"b\",\n        {\n            \"my_integer\": 2,\n            \"my_string\": \"c\",\n            \"my_list\": [3,\"d\"]\n        }\n    ],\n    \"my_dict\": {\n        \"my_integer\": 4,\n        \"my_string\": \"e\",\n        \"my_list\": [5,\"f\"],\n        \"my_dict\": {\n            \"my_integer\": 6,\n            \"my_string\": \"g\"\n        }\n    },\n    \"my_list_empty\": [],\n    \"my_dict_empty\": {},\n    \"my_subclass\": {\n        \"my_integer\": 7,\n        \"my_string\": \"h\",\n        \"my_list\": [8,\"i\"],\n        \"my_dict\": {\n            \"my_integer\": 9,\n            \"my_string\": \"j\"\n        }\n    }\n}"  # noqa: E501
        self.assertEqual(mcs.Stringifier.class_to_str(MyClass()), expected)

    def test_class_to_str_with_empty_class(self):
        self.assertEqual(mcs.Stringifier.class_to_str(MyEmptyclass()), "{}")

    def test_generate_pretty_object_output(self):
        object_list = [
            mcs.ObjectMetadata(
                uuid='id1',
                shape='',
                state_list=[],
                texture_color_list=[],
                held=True,
                visible=True,
                position=None,
                dimensions=None,
                distance_in_world=0,
                direction=None
            ),
            mcs.ObjectMetadata(
                uuid='really_long_id2',
                shape='sofa',
                state_list=['state1', 'state2'],
                texture_color_list=['black', 'white'],
                held=False,
                visible=False,
                position={
                    'x': 1,
                    'y': 2,
                    'z': 3
                },
                dimensions=[{
                    'x': 4,
                    'y': 5,
                    'z': 6
                }],
                distance_in_world=1234567890987654321,
                direction={
                    'x': 10000,
                    'y': 20000,
                    'z': 30000
                }
            )
        ]
        self.assertEqual(mcs.Stringifier.generate_pretty_object_output(
            object_list), [
            'OBJECT ID        SHAPE  COLORS        HELD   VISIBLE  STATE           POSITION (WORLD)  DISTANCE (WORLD)     DIRECTION (WORLD)    DIMENSIONS (WORLD)',  # noqa: E501
            'id1                                   True   True                     None              0                    None                 None              ',  # noqa: E501
            'really_long_id2  sofa   black, white  False  False    state1, state2  (1,2,3)           1234567890987654321  (10000,20000,30000)  [(4,5,6)]         '  # noqa: E501
        ])

    def test_value_to_str_with_boolean(self):
        self.assertEqual(mcs.Stringifier.value_to_str(True), "true")
        self.assertEqual(mcs.Stringifier.value_to_str(False), "false")

    def test_value_to_str_with_dict(self):
        self.assertEqual(mcs.Stringifier.value_to_str({}), "{}")
        self.assertEqual(mcs.Stringifier.value_to_str({
            "number": 1,
            "string": "a"
        }), "{\n    \"number\": 1,\n    \"string\": \"a\"\n}")

    def test_value_to_str_with_float(self):
        self.assertEqual(mcs.Stringifier.value_to_str(0.0), "0.0")
        self.assertEqual(mcs.Stringifier.value_to_str(1234.5678), "1234.5678")
        self.assertEqual(mcs.Stringifier.value_to_str(0.12345678), "0.123457")
        self.assertEqual(
            mcs.Stringifier.value_to_str(-0.12345678),
            "-0.123457"
        )

    def test_value_to_str_with_integer(self):
        self.assertEqual(mcs.Stringifier.value_to_str(0), "0")
        self.assertEqual(mcs.Stringifier.value_to_str(1234), "1234")

    def test_value_to_str_with_list(self):
        self.assertEqual(mcs.Stringifier.value_to_str([]), "[]")
        self.assertEqual(mcs.Stringifier.value_to_str([1, "a"]), "[1,\"a\"]")

    def test_value_to_str_with_list_with_nested_dict(self):
        self.assertEqual(mcs.Stringifier.value_to_str([]), "[]")
        self.assertEqual(
            mcs.Stringifier.value_to_str([1, "a", {"b": 2}]),
            "[\n    1,\n    \"a\",\n    {\n        \"b\": 2\n    }\n]"
        )

    def test_value_to_str_with_list_with_nested_list(self):
        self.assertEqual(mcs.Stringifier.value_to_str([]), "[]")
        self.assertEqual(
            mcs.Stringifier.value_to_str([1, "a", [2, "b"]]),
            "[\n    1,\n    \"a\",\n    [2,\"b\"]\n]"
        )

    def test_value_to_str_with_string(self):
        self.assertEqual(mcs.Stringifier.value_to_str(""), "\"\"")
        self.assertEqual(
            mcs.Stringifier.value_to_str("a b c d"),
            "\"a b c d\"")

    def test_vector_to_string(self):
        self.assertEqual(mcs.Stringifier.vector_to_string(None), 'None')
        self.assertEqual(mcs.Stringifier.vector_to_string({
            'x': 1,
            'y': 2,
            'z': 3
        }), '(1,2,3)')


if __name__ == '__main__':
    unittest.main()



import concurrent.futures
import json
import logging
import logging.config
import signal
from contextlib import contextmanager
from typing import Dict, Union

import typeguard

from ._version import __version__
from .action import Action
from .config_manager import ConfigManager
from .controller import Controller
from .goal_metadata import GoalCategory, GoalMetadata
from .history_writer import HistoryWriter
from .logging_config import LoggingConfig
from .material import Material
from .object_metadata import ObjectMetadata
from .return_status import ReturnStatus
from .reward import Reward
from .scene_history import SceneHistory
from .serializer import SerializerMsgPack
from .step_metadata import StepMetadata
from .stringifier import Stringifier
from .subscriber import add_subscribers
from .unity_executable_provider import UnityExecutableProvider

logger = logging.getLogger(__name__)
# Set default logging handler to avoid "No handler found" warnings
logger.addHandler(logging.NullHandler())

# Timeout at 3 minutes (180 seconds).  It was 60 seconds but
# this can cause timeouts on EC2 instances
TIME_LIMIT_SECONDS = 180


def get_controller(unity_exec: str, config: ConfigManager):
    """Function to get the controller, pulled into its own
     function so we can time it.  """
    controller = Controller(unity_exec, config)
    return controller


def get_controller_with_timeout(unity_exec: str, config: ConfigManager):
    """Wrapper function that sets a timeout for the controller creation.
    If getting the controller times out, None is returned. """
    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        future = executor.submit(get_controller, unity_exec, config)
        try:
            controller = future.result(timeout=TIME_LIMIT_SECONDS)
            return controller
        except concurrent.futures.TimeoutError as Msg:
            logger.error("Timeout error in creating controller", exc_info=Msg)

            # TODO:  Add checks to continue waiting for the controller creation.
            return None


@typeguard.typechecked
def init_logging(log_config: Dict = None,
                 log_config_file: str = "log.config.user.py"):
    """
    Initializes logging system.  If no parameters are provided, a
    default configuration will be applied.  See python logging
    documentation for details.

    https://docs.python.org/3/library/logging.config.html#logging-config-dictschema

    Parameters
    ----------
    log_config : dict, optional
        A dictionary the contains the logging configuration.  If None, a
        default configuration will be used
    log_config_file: str, optional
        Path to an override configuration file.  The file will contain a
        python dictionary for the logging configuration.  This file is
        typically not used, but allows a user to change the logging
        configuration without code changes.  Default, log.config.user.py
    """
    LoggingConfig.init_logging(
        log_config=log_config,
        log_config_file=log_config_file)


@typeguard.typechecked
def create_controller(config_file_or_dict: Union[Dict, str] = None,
                      unity_app_file_path: str = None,
                      unity_cache_version: str = None):
    """
    Creates and returns a new MCS Controller object.

    Parameters
    ----------
    config_file_or_dict: str or dict, required
        Can be a path to configuration file to read in or a dictionary
        of various properties, such as metadata level and whether or
        not to save history files (default None)

        * Note the **order of precedence for config options**, in case more
          than one is given:

        1. **MCS_CONFIG_FILE_PATH** environment variable (meant for internal
           TA2 use during evaluation)
        2. If no environment variable given, use **config_file_or_dict**
           parameter. The value can be a string file path or a dictionary.
        3. Raises FileNotFoundError if no config found.
    unity_app_file_path : str, optional
        The file path to your MCS Unity application.  If Not provided,
        the internal cache and downloader will attempt to locate and use
        the current version.
        (default None)
    unity_cache_version : str, optional
        If no file path is provided for the MCS Unity application, the
        version provided will be found via cache and internal downloader.
        If not provided, the version matching the MCS code will be used.
        (default None)

    Returns
    -------
    Controller
        The MCS Controller object.
    """
    try:
        unity_exec = unity_app_file_path
        if (unity_exec is None):
            unity_provider = UnityExecutableProvider()
            unity_exec = unity_provider.get_executable(
                unity_cache_version).as_posix()

        config = ConfigManager(config_file_or_dict)
        controller = get_controller_with_timeout(unity_exec, config)
        if not controller:
            raise Exception('MCS/Python Controller failed to initialize')
        add_subscribers(controller, config)
        return controller
    except Exception as Msg:
        logger.error("Exception in create_controller()", exc_info=Msg)
        return None


@typeguard.typechecked
def change_config(controller: Controller,
                  config_file_or_dict: Union[Dict, str] = None):
    """
    Creates and returns a new MCS Controller object.  Should only be called
    After a run and before a scene is changed.

    Parameters
    ----------
    controller : Controller
        The currently used controller that the config should be changed
        on.
    config_file_or_dict: str or dict, optional
        Can be a path to configuration file to read in or a dictionary
        of various properties, such as metadata level and whether or
        not to save history files (default None)

    """
    config = ConfigManager(config_file_or_dict)
    controller._set_config(config)
    controller.remove_all_event_handlers()
    add_subscribers(controller, config)


@typeguard.typechecked
def load_scene_json_file(scene_json_file_path: str) -> Dict:
    """
    Loads the given JSON scene config file and returns its data.

    Parameters
    ----------
    config_json_file_path : str
        The file path to your MCS JSON scene configuration file.

    Returns
    -------
    dict
        The MCS scene configuration data from the given JSON file.

    Raises
    ------
    FileNotFoundError
    ValueError
    """
    with open(scene_json_file_path, encoding='utf-8-sig') \
            as config_json_file_object:
        return json.load(config_json_file_object)



import io
from abc import ABCMeta, abstractmethod

import msgpack
import numpy as np
import PIL.Image as Image

from .goal_metadata import GoalMetadata
from .object_metadata import ObjectMetadata
from .step_metadata import StepMetadata


class ISerializer:
    __metaclass__ = ABCMeta

    @classmethod
    def version(cls):
        return "0.1.0"

    @staticmethod
    @abstractmethod
    def serialize(unpacked_object):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def deserialize(packed_object):
        raise NotImplementedError

    @staticmethod
    def image_to_bytes(image):
        """Converts PIL image to bytes."""
        byte_io = io.BytesIO()
        image.save(byte_io, format='PNG')
        return byte_io.getvalue()

    @staticmethod
    def bytes_to_image(bytes_image):
        """Converts bytes to PIL image."""
        return Image.open(io.BytesIO(bytes_image))


class SerializerMsgPack(ISerializer):
    """Serializer to (de)serialize StepMetadata into/from MsgPack format."""

    @staticmethod
    def _ext_pack(x):
        """Hook to serialize MCS Step Metadata as MsgPack, e.g.
        serialized = msgpack.packb(output, default=ext_pack, strict_types=True)
        """
        if isinstance(x, StepMetadata):
            # TODO In future can investigate reflection for automating field
            # extraction, but notice that strict types flag is highly
            # dependent on order. Something like this might be a start:
            # [field for field in dir(x) if not callable(getattr(x, field))
            # and not field.startswith("__")]
            return msgpack.ExtType(
                1,
                msgpack.packb([
                    x.action_list, x.camera_aspect_ratio,
                    x.camera_clipping_planes, x.camera_field_of_view,
                    x.camera_height, x.depth_map_list, x.goal, x.head_tilt,
                    x.image_list, x.object_list, x.object_mask_list,
                    x.position, x.return_status, x.reward, x.rotation,
                    x.step_number, x.structural_object_list
                ],
                    default=SerializerMsgPack._ext_pack,
                    strict_types=True))
        elif isinstance(x, tuple):
            return msgpack.ExtType(
                2,
                msgpack.packb([x[0], x[1]],
                              default=SerializerMsgPack._ext_pack,
                              strict_types=True))
        elif isinstance(x, Image.Image):
            return msgpack.ExtType(
                3,
                msgpack.packb(SerializerMsgPack.image_to_bytes(x),
                              default=SerializerMsgPack._ext_pack,
                              strict_types=True))
        elif isinstance(x, GoalMetadata):
            return msgpack.ExtType(
                4,
                msgpack.packb([
                    x.action_list, x.category, x.description,
                    x.habituation_total, x.last_preview_phase_step,
                    x.last_step, x.metadata
                ],
                    default=SerializerMsgPack._ext_pack,
                    strict_types=True))
        elif isinstance(x, ObjectMetadata):
            return msgpack.ExtType(
                5,
                msgpack.packb([
                    x.uuid, x.dimensions, x.direction, x.distance,
                    x.distance_in_steps, x.distance_in_world, x.held, x.mass,
                    x.material_list, x.position, x.rotation, x.segment_color,
                    x.shape, x.state_list, x.texture_color_list, x.visible
                ],
                    default=SerializerMsgPack._ext_pack,
                    strict_types=True))
        elif isinstance(x, np.ndarray):
            return msgpack.ExtType(
                6,
                msgpack.packb(x.tolist(),
                              default=SerializerMsgPack._ext_pack,
                              strict_types=True))
        return x

    @staticmethod
    def _ext_unpack(code, data):
        """
        Hook to deserialize MCS Step Metadata from MsgPack, e.g.
        deserialized = msgpack.unpackb(packed_bytes, ext_hook=ext_unpack)
        """
        if code == 1:
            action_list, camera_aspect_ratio, camera_clipping_planes, \
                camera_field_of_view, camera_height, depth_map_list, goal, \
                head_tilt, image_list, object_list, object_mask_list, \
                position, return_status, reward, rotation, step_number, \
                structural_object_list = \
                msgpack.unpackb(data, ext_hook=SerializerMsgPack._ext_unpack)
            return StepMetadata(action_list=action_list,
                                camera_aspect_ratio=camera_aspect_ratio,
                                camera_clipping_planes=camera_clipping_planes,
                                camera_field_of_view=camera_field_of_view,
                                camera_height=camera_height,
                                depth_map_list=depth_map_list,
                                goal=goal,
                                head_tilt=head_tilt,
                                image_list=image_list,
                                object_list=object_list,
                                object_mask_list=object_mask_list,
                                position=position,
                                return_status=return_status,
                                reward=reward,
                                rotation=rotation,
                                step_number=step_number,
                                structural_object_list=structural_object_list)
        elif code == 2:
            x0, x1 = msgpack.unpackb(data,
                                     ext_hook=SerializerMsgPack._ext_unpack)
            return x0, x1
        elif code == 3:
            x = msgpack.unpackb(data, ext_hook=SerializerMsgPack._ext_unpack)
            return SerializerMsgPack.bytes_to_image(x)
        elif code == 4:
            action_list, category, description, habituation_total, \
                last_preview_phase_step, last_step, metadata = \
                msgpack.unpackb(data, ext_hook=SerializerMsgPack._ext_unpack)
            return GoalMetadata(action_list, category, description,
                                habituation_total, last_preview_phase_step,
                                last_step, metadata)
        elif code == 5:
            uuid, dimensions, direction, distance, distance_in_steps, \
                distance_in_world, held, mass, material_list, position, \
                rotation, segment_color, shape, state_list, \
                texture_color_list, visible = msgpack.unpackb(
                    data, ext_hook=SerializerMsgPack._ext_unpack)
            return ObjectMetadata(uuid, dimensions, direction, distance,
                                  distance_in_steps, distance_in_world, held,
                                  mass, material_list, position, rotation,
                                  segment_color, shape, state_list,
                                  texture_color_list, visible)
        elif code == 6:
            x = msgpack.unpackb(data, ext_hook=SerializerMsgPack._ext_unpack)
            return np.asarray(x)
        return msgpack.ExtType(code, data)

    @staticmethod
    def serialize(step_metadata: StepMetadata):
        """
        Serializes step metadata into MsgPack.

        You can use
        .. code-block:: python

            object_to_persist = {
                'payload': step_metadata,
                'additional_info': 'info'
                }

        to add extra data.

        Args:
            step_metadata: MCS step metadata output

        Returns:
            Serialized version of step metadata in MsgPack format.
        """
        return msgpack.packb(step_metadata,
                             default=SerializerMsgPack._ext_pack,
                             strict_types=True)

    @staticmethod
    def deserialize(packed_step_metadata):
        return msgpack.unpackb(packed_step_metadata,
                               ext_hook=SerializerMsgPack._ext_unpack)



import copy
import glob
import os
import shutil
import unittest

import numpy as np

import machine_common_sense as mcs
from machine_common_sense.config_manager import SceneConfiguration

TEST_FILE_NAME = "test_scene_file.json"
PREFIX = 'prefix'


class TestHistoryWriter(unittest.TestCase):

    config_data = SceneConfiguration(name=TEST_FILE_NAME)
    prefix_config_data = SceneConfiguration(name=f"{PREFIX}/{TEST_FILE_NAME}")

    @classmethod
    def tearDownClass(cls):
        # remove all TEST_FILE_NAME in PREFIX
        test_file_base = os.path.splitext(TEST_FILE_NAME)[0]
        prefix_test_files = glob.glob(
            f'{mcs.HistoryWriter.HISTORY_DIRECTORY}/{PREFIX}/{test_file_base}*'
        )
        for prefix_test_file in prefix_test_files:
            os.unlink(prefix_test_file)
        # if PREFIX empty, destroy it
        try:
            if not os.listdir(
                    f"{mcs.HistoryWriter.HISTORY_DIRECTORY}/{PREFIX}"):
                shutil.rmtree(
                    f"{mcs.HistoryWriter.HISTORY_DIRECTORY}/{PREFIX}")
        except BaseException:
            pass

        # remove all TEST_FILE_NAME in SCENE_HIST_DIR
        test_files = glob.glob(
            f'{mcs.HistoryWriter.HISTORY_DIRECTORY}/{test_file_base}*')
        for test_file in test_files:
            os.unlink(test_file)
        # if SCENE_HIST_DIR is empty, destroy it
        if not os.listdir(mcs.HistoryWriter.HISTORY_DIRECTORY):
            shutil.rmtree(mcs.HistoryWriter.HISTORY_DIRECTORY)

    def test_init(self):
        writer = mcs.HistoryWriter(self.config_data)

        self.assertEqual(writer.info_obj.keys(), {'name', 'timestamp'})
        self.assertEqual(writer.info_obj['name'], "test_scene_file")
        self.assertEqual(writer.info_obj['timestamp'], "")
        self.assertTrue(os.path.exists(writer.HISTORY_DIRECTORY))

    def test_init_with_hist_info(self):
        config_data = SceneConfiguration(name=TEST_FILE_NAME)
        writer = mcs.HistoryWriter(config_data, {
            'team': 'team1',
            'metadata': 'level1'
        })

        self.assertEqual(writer.info_obj.keys(), {
            'team',
            'metadata',
            'name',
            'timestamp'
        })
        self.assertEqual(writer.info_obj['name'], "test_scene_file")
        self.assertEqual(writer.info_obj['team'], "team1")
        self.assertEqual(writer.info_obj['metadata'], "level1")
        self.assertEqual(writer.info_obj['timestamp'], "")
        self.assertTrue(os.path.exists(writer.HISTORY_DIRECTORY))

    def test_init_with_timestamp(self):
        writer = mcs.HistoryWriter(self.config_data, {}, "some-time-value")

        self.assertEqual(writer.info_obj.keys(), {'name', 'timestamp'})
        self.assertEqual(writer.info_obj['name'], "test_scene_file")
        self.assertEqual(writer.info_obj['timestamp'], "some-time-value")
        self.assertTrue(os.path.exists(writer.HISTORY_DIRECTORY))

    def test_add_step(self):
        writer = mcs.HistoryWriter(self.config_data)

        history_item = mcs.SceneHistory(
            step=1,
            action=mcs.Action.MOVE_AHEAD.value)
        writer.add_step(history_item)

        self.assertEqual(len(writer.current_steps), 1)
        self.assertEqual(
            writer.current_steps[0]["action"],
            mcs.Action.MOVE_AHEAD.value)

        history_item = mcs.SceneHistory(
            step=2,
            action=mcs.Action.MOVE_LEFT.value)
        writer.add_step(history_item)

        self.assertEqual(len(writer.current_steps), 2)
        self.assertEqual(
            writer.current_steps[1]["action"],
            mcs.Action.MOVE_LEFT.value)

    def test_add_step_timer(self):
        writer = mcs.HistoryWriter(self.config_data)
        writer.init_timer()
        # set start time back 500ms as if the run took 500ms.
        writer.last_step_time_millis -= 500
        # Save the time so we can check that it changes
        initial_time = writer.last_step_time_millis
        history_item = mcs.SceneHistory(
            step=1,
            action=mcs.Action.MOVE_AHEAD.value)
        writer.add_step(history_item)

        self.assertEqual(len(writer.current_steps), 1)
        # we give some delta here because commands do take some time to run
        self.assertAlmostEqual(
            writer.current_steps[0]["delta_time_millis"], 500, delta=1)
        self.assertNotEqual(initial_time, writer.last_step_time_millis)

        writer.last_step_time_millis -= 300
        history_item = mcs.SceneHistory(
            step=2,
            action=mcs.Action.MOVE_LEFT.value)
        writer.add_step(history_item)

        self.assertEqual(len(writer.current_steps), 2)
        self.assertAlmostEqual(
            writer.current_steps[1]["delta_time_millis"], 300, delta=1)

        history_item = mcs.SceneHistory(
            step=3,
            action=mcs.Action.MOVE_LEFT.value)
        writer.add_step(history_item)

        self.assertEqual(len(writer.current_steps), 3)
        self.assertAlmostEqual(
            writer.current_steps[2]["delta_time_millis"], 0, delta=1)

    def test_write_history_file(self):
        writer = mcs.HistoryWriter(self.config_data)

        history_item = mcs.SceneHistory(
            step=1,
            action=mcs.Action.MOVE_AHEAD.value)
        writer.add_step(history_item)

        history_item = mcs.SceneHistory(
            step=2,
            action=mcs.Action.MOVE_LEFT.value)
        writer.add_step(history_item)

        writer.write_history_file("Plausible", 0.75)

        self.assertEqual(writer.end_score["classification"], "Plausible")
        self.assertEqual(writer.end_score["confidence"], "0.75")

        self.assertEqual(writer.history_obj["info"]["name"], "test_scene_file")
        self.assertEqual(len(writer.history_obj["steps"]), 2)
        self.assertEqual(
            writer.history_obj["score"]["classification"], "Plausible")
        self.assertEqual(writer.history_obj["score"]["confidence"], "0.75")

        self.assertTrue(os.path.exists(writer.scene_history_file))

    def test_write_history_file_with_slash(self):
        writer = mcs.HistoryWriter(self.prefix_config_data)

        history_item = mcs.SceneHistory(
            step=1,
            action=mcs.Action.MOVE_AHEAD.value)
        writer.add_step(history_item)

        history_item = mcs.SceneHistory(
            step=2,
            action=mcs.Action.MOVE_LEFT.value)
        writer.add_step(history_item)

        writer.write_history_file("Plausible", 0.75)

        self.assertEqual(writer.end_score["classification"], "Plausible")
        self.assertEqual(writer.end_score["confidence"], "0.75")

        self.assertEqual(
            writer.history_obj["info"]["name"],
            "prefix/test_scene_file")
        self.assertEqual(len(writer.history_obj["steps"]), 2)
        self.assertEqual(
            writer.history_obj["score"]["classification"], "Plausible")
        self.assertEqual(writer.history_obj["score"]["confidence"], "0.75")

        self.assertTrue(os.path.exists(writer.scene_history_file))

    def test_write_step_removes_some_output(self):
        writer = mcs.HistoryWriter(self.prefix_config_data)

        output = mcs.StepMetadata(
            action_list=[
                mcs.Action.CLOSE_OBJECT.value,
                mcs.Action.MOVE_AHEAD.value],
            return_status="SUCCESSFUL",
            step_number=2,
            object_list={
                "object": "object"
            },
            structural_object_list={
                "structural_object": "structural_object"
            }
        )

        history_item = mcs.SceneHistory(
            step=1,
            action=mcs.Action.MOVE_AHEAD.value,
            output=copy.deepcopy(output)
        )
        writer.add_step(history_item)

        self.assertEqual(len(writer.current_steps), 1)
        self.assertEqual(
            writer.current_steps[0]["action"],
            mcs.Action.MOVE_AHEAD.value)
        self.assertIsNone(writer.current_steps[0]["output"]["action_list"])
        self.assertIsNone(writer.current_steps[0]["output"]["object_list"])
        self.assertIsNone(
            writer.current_steps[0]["output"]["structural_object_list"])

        history_item = mcs.SceneHistory(
            step=2,
            action=mcs.Action.MOVE_LEFT.value,
            output=copy.deepcopy(output)
        )
        writer.add_step(history_item)

        self.assertEqual(len(writer.current_steps), 2)
        self.assertEqual(
            writer.current_steps[1]["action"],
            mcs.Action.MOVE_LEFT.value)
        self.assertIsNone(writer.current_steps[1]["output"]["action_list"])
        self.assertIsNone(writer.current_steps[1]["output"]["object_list"])
        self.assertIsNone(
            writer.current_steps[1]["output"]["structural_object_list"])

    def test_write_step_has_target_updates_some_output(self):
        writer = mcs.HistoryWriter(self.prefix_config_data)

        goal = mcs.GoalMetadata(metadata={
            'target': {'id': 'targetId', 'image': 'something.png'},
            'targets': [{'id': 'targetId', 'image': 'something.png'}]
        })
        output = mcs.StepMetadata(
            action_list=[
                mcs.Action.CLOSE_OBJECT.value,
                mcs.Action.MOVE_AHEAD.value],
            return_status="SUCCESSFUL",
            step_number=1,
            object_list=[
                mcs.ObjectMetadata(
                    uuid="targetId", position={
                        "x": 1, "y": 0.5, "z": 1})
            ],
            structural_object_list=[
                mcs.ObjectMetadata(uuid='struct_obj')
            ],
            goal=goal
        )

        history_item = mcs.SceneHistory(
            step=1,
            action=mcs.Action.MOVE_AHEAD.value,
            output=copy.deepcopy(output)
        )
        writer.add_step(history_item)

        self.assertEqual(len(writer.current_steps), 1)
        self.assertEqual(
            writer.current_steps[0]["action"],
            mcs.Action.MOVE_AHEAD.value)
        self.assertIsNone(writer.current_steps[0]["output"]["action_list"])
        self.assertIsNone(writer.current_steps[0]["output"]["object_list"])
        self.assertIsNone(
            writer.current_steps[0]["output"]["structural_object_list"])
        self.assertEqual(
            writer.current_steps[0]["output"]["goal"]["metadata"]["target"],
            {'id': 'targetId', 'position': {'x': 1, 'y': 0.5, 'z': 1}})
        self.assertEqual(
            writer.current_steps[0]["output"]["goal"]["metadata"]["targets"],
            [{'id': 'targetId', 'position': {'x': 1, 'y': 0.5, 'z': 1}}])

    def test_write_history_file_with_numpy(self):
        writer = mcs.HistoryWriter(self.prefix_config_data)

        history_item = mcs.SceneHistory(
            step=np.int32(1),
            action=mcs.Action.MOVE_AHEAD.value)
        writer.add_step(history_item)

        history_item = mcs.SceneHistory(
            step=2,
            action=mcs.Action.MOVE_LEFT.value)
        writer.add_step(history_item)

        writer.write_history_file("Plausible", np.float64(0.75))

        self.assertEqual(writer.end_score["classification"], "Plausible")
        self.assertEqual(writer.end_score["confidence"], "0.75")

        self.assertEqual(
            writer.history_obj["info"]["name"],
            "prefix/test_scene_file")
        self.assertEqual(len(writer.history_obj["steps"]), 2)
        self.assertEqual(
            writer.history_obj["score"]["classification"], "Plausible")
        self.assertEqual(writer.history_obj["score"]["confidence"], "0.75")

        self.assertTrue(os.path.exists(writer.scene_history_file))

    def test_is_target_visible_retrieval(self):
        writer = mcs.HistoryWriter(self.prefix_config_data)

        history_1 = mcs.SceneHistory(output=mcs.StepMetadata(
            goal=mcs.GoalMetadata(category='retrieval', metadata={
                'target': {'id': 'target_1'}
            }),
            object_list=[
                mcs.ObjectMetadata(uuid='target_1', visible=False)
            ]
        ))
        actual = writer.is_target_visible(history_1)
        self.assertFalse(actual)

        history_2 = mcs.SceneHistory(output=mcs.StepMetadata(
            goal=mcs.GoalMetadata(category='retrieval', metadata={
                'target': {'id': 'target_1'}
            }),
            object_list=[
                mcs.ObjectMetadata(uuid='target_1', visible=True)
            ]
        ))
        actual = writer.is_target_visible(history_2)
        self.assertTrue(actual)

    def test_is_target_visible_multi_retrieval(self):
        writer = mcs.HistoryWriter(self.prefix_config_data)

        history_1 = mcs.SceneHistory(output=mcs.StepMetadata(
            goal=mcs.GoalMetadata(category='multi retrieval', metadata={
                'targets': [{'id': 'target_1'}]
            }),
            object_list=[
                mcs.ObjectMetadata(uuid='target_1', visible=False)
            ]
        ))
        actual = writer.is_target_visible(history_1)
        self.assertEqual(actual, [])

        history_2 = mcs.SceneHistory(output=mcs.StepMetadata(
            goal=mcs.GoalMetadata(category='multi retrieval', metadata={
                'targets': [{'id': 'target_1'}]
            }),
            object_list=[
                mcs.ObjectMetadata(uuid='target_1', visible=True)
            ]
        ))
        actual = writer.is_target_visible(history_2)
        self.assertEqual(actual, ['target_1'])

        history_3 = mcs.SceneHistory(output=mcs.StepMetadata(
            goal=mcs.GoalMetadata(category='multi retrieval', metadata={
                'targets': [{'id': 'target_1'}, {'id': 'target_2'}]
            }),
            object_list=[
                mcs.ObjectMetadata(uuid='target_1', visible=True),
                mcs.ObjectMetadata(uuid='target_2', visible=False)
            ]
        ))
        actual = writer.is_target_visible(history_3)
        self.assertEqual(actual, ['target_1'])

        history_4 = mcs.SceneHistory(output=mcs.StepMetadata(
            goal=mcs.GoalMetadata(category='multi retrieval', metadata={
                'targets': [{'id': 'target_1'}, {'id': 'target_2'}]
            }),
            object_list=[
                mcs.ObjectMetadata(uuid='target_1', visible=True),
                mcs.ObjectMetadata(uuid='target_2', visible=True)
            ]
        ))
        actual = writer.is_target_visible(history_4)
        self.assertEqual(actual, ['target_1', 'target_2'])


if __name__ == '__main__':
    unittest.main()
